# Generated by Django 5.0.12 on 2025-09-08 11:39

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('wbfdm', '0032_alter_instrumentprice_outstanding_shares'),
    ]

    operations = [
        migrations.AlterField(
            model_name='controversy',
            name='review',
            field=models.DateField(blank=True, null=True, verbose_name='Reviewed'),
        ),
    ]



from django.db.models.signals import ModelSignal

# this signal is triggered whenever all instruments metrics are updated. Temporary solution until we rework the framework for more dynamicity

instrument_metric_updated = ModelSignal(
    use_caching=True
)  # the sender model is the type model class being updated (e.g. Instrument), expect a "basket", "key" and "val_date" keyword argument (null if all are updated)



from contextlib import suppress
from typing import Optional

from django.utils.translation import gettext as _
from dynamic_preferences.registries import global_preferences_registry
from rest_framework.reverse import reverse
from wbcore.metadata.configs import display as dp
from wbcore.metadata.configs.display import DisplayViewConfig
from wbcore.metadata.configs.display.instance_display.layouts.inlines import Inline
from wbcore.metadata.configs.display.instance_display.layouts.layouts import Layout
from wbcore.metadata.configs.display.instance_display.pages import Page
from wbcore.metadata.configs.display.instance_display.shortcuts import (
    Display,
    create_simple_page_with_inline,
    create_simple_section,
    default,
)
from wbcore.metadata.configs.display.instance_display.styles import Style
from wbcore.metadata.configs.display.instance_display.utils import repeat, repeat_field

from wbfdm.contrib.metric.viewsets.configs.utils import (
    get_performance_fields,
    get_statistic_field,
)
from wbfdm.models import Instrument

ALL_GRIDS = [
    "swe-summary",
    "swe-income-statement",
    "swe-balance-sheet",
    "swe-cashflow-statement",
    "swe-ratios",
    "swe-margins",
    "swe-cashflow-ratios",
    "swe-asset-turnover-ratios",
    "swe-credit",
    "swe-long-term-solvency",
    "swe-short-term-liquidity",
]


def _get_hierarchy_section_title(instrument_id):
    with suppress(Instrument.DoesNotExist):
        instrument = Instrument.objects.get(id=instrument_id)
        if instrument.children.exists():
            if instrument.level == 0:
                return "Securities"
            elif instrument.level == 1:
                return "Quotes"
    return "Children"


def get_financial_summary_section_name() -> str:
    try:
        return global_preferences_registry.manager()["wbfdm__financial_summary_section_name"]
    except Exception:
        return "Financial Summary"


def remove_ref_from_list(grid_list: list, ref: str) -> list[str]:
    _l = grid_list.copy()
    _l.remove(ref)
    return _l


class InstrumentDisplayConfig(DisplayViewConfig):
    def get_list_display(self) -> Optional[dp.ListDisplay]:
        return dp.ListDisplay(
            fields=[
                dp.Field(key="name", label=_("Name"), pinned="left"),
                dp.Field(
                    key=None,
                    label=_("Information"),
                    open_by_default=False,
                    children=[
                        dp.Field(key="instrument_type", label=_("Instrument Type"), show="open"),
                        dp.Field(key="exchange", label=_("Exchange")),
                        dp.Field(key="isin", label=_("ISIN")),
                        dp.Field(key="ticker", label=_("Ticker")),
                        dp.Field(key="refinitiv_identifier_code", label=_("RIC")),
                        dp.Field(key="refinitiv_mnemonic_code", label=_("Refinitiv Mnemonic")),
                        dp.Field(key="description", label=_("Description"), show="open"),
                        dp.Field(key="currency", label=_("Currency"), show="open"),
                        dp.Field(key="country", label=_("Country"), show="open"),
                        dp.Field(key="classifications", label="Classifications", show="open"),
                        dp.Field(key="tags", label="Tags", show="open"),
                    ],
                ),
                dp.Field(
                    key=None,
                    label=_("Extra"),
                    open_by_default=False,
                    children=[
                        dp.Field(key="is_primary", label=_("Primary")),
                        dp.Field(key="is_preferred_primary", label=_("Preferred Primary"), show="open"),
                        dp.Field(key="is_investable_universe", label=_("Investable Universe"), show="open"),
                        dp.Field(key="is_security", label=_("Security"), show="open"),
                        dp.Field(key="is_managed", label=_("Internally Managed"), show="open"),
                    ],
                ),
                dp.Field(key=None, label=_("Performance"), open_by_default=False, children=get_performance_fields()),
                get_statistic_field(),
            ],
            tree=True,
            tree_group_field="name",
            tree_group_level_options=[
                dp.TreeGroupLevelOption(
                    filter_key="parent",
                    filter_depth=1,
                    filter_blacklist=["parent__isnull"],
                    list_endpoint=reverse(
                        "wbfdm:instrument-list",
                        args=[],
                        request=self.request,
                    ),
                )
            ],
        )

    def get_instance_display(self):
        return Display(
            pages=[
                Page(
                    title="Home",
                    layouts={
                        default(): Layout(
                            grid_template_areas=[
                                ["name", "ticker", "isin", "exchange", "market_data_chart", "valuation_ratios-new"],
                                [
                                    "name_repr",
                                    "instrument_type",
                                    "currency",
                                    "country",
                                    "market_data_chart",
                                    "valuation_ratios-new",
                                ],
                                [
                                    "inception_date",
                                    "delisted_date",
                                    ".",
                                    ".",
                                    "market_data_chart",
                                    "valuation_ratios-new",
                                ],
                                [
                                    "description",
                                    "description",
                                    "description",
                                    "description",
                                    "market_data_chart",
                                    "valuation_ratios-new",
                                ],
                                [repeat_field(4, "fin-summary"), repeat_field(2, ".")],
                            ],
                            grid_template_rows=["min-content"] * 3 + ["1fr", "446px"],
                            grid_template_columns=[repeat(4, "183px"), "1fr", "1fr"],
                            inlines=[
                                Inline(key="market_data_chart", endpoint="market_data", title="Prices"),
                                Inline(key="valuation_ratios-new", endpoint="valuation_ratios-new", title="Ratios"),
                                Inline(
                                    key="fin-summary",
                                    endpoint="fin-summary",
                                    title="Financial Summary",
                                    hide_controls=True,
                                ),
                            ],
                        ),
                    },
                ),
                Page(
                    title="Fundamentals",
                    display=Display(
                        navigation_type=dp.NavigationType.PANEL,
                        pages=[
                            create_simple_page_with_inline("Summary", "swe-summary"),
                            create_simple_page_with_inline("Income Statement", "swe-income-statement"),
                            create_simple_page_with_inline("Balance Sheet", "swe-balance-sheet"),
                            create_simple_page_with_inline("Cashflow Statement", "swe-cashflow-statement"),
                            create_simple_page_with_inline("Ratios", "swe-ratios"),
                        ],
                    ),
                ),
                Page(
                    title="Performance",
                    display=Display(
                        navigation_type=dp.NavigationType.PANEL,
                        pages=[
                            dp.Page(
                                title="Overview",
                                layouts={
                                    default(): dp.Layout(
                                        grid_template_areas=[
                                            ["market_data", "market_data", "market_data"],
                                            ["bestandworstreturns", "performance_summary", "."],
                                        ],
                                        grid_template_rows=["450px", "300px"],
                                        grid_template_columns=["2fr", "2fr", "1fr"],
                                        inlines=[
                                            dp.Inline(key="market_data", endpoint="market_data"),
                                            dp.Inline(
                                                key="performance_summary",
                                                endpoint="performance_summary",
                                                hide_controls=True,
                                                title="Performance Summary",
                                            ),
                                            dp.Inline(
                                                key="bestandworstreturns",
                                                endpoint="bestandworstreturns",
                                                hide_controls=True,
                                                title="Best / Worst Returns",
                                            ),
                                        ],
                                    )
                                },
                            ),
                            create_simple_page_with_inline("Prices", "prices"),
                            create_simple_page_with_inline("Statistics", "financial-statistics"),
                        ],
                    ),
                ),
                Page(
                    title="Classifications & Lists",
                    layouts={
                        default(): Layout(
                            grid_template_areas=[
                                ["classifications_list", "instrument_lists"],
                            ],
                            grid_template_columns=[Style.pct(50), Style.pct(50)],
                            sections=[
                                create_simple_section(
                                    key="classifications_list",
                                    title="Classifications",
                                    grid_template_areas=[["classifications_list"]],
                                    inline_key="classifications_list",
                                    collapsible=False,
                                ),
                                create_simple_section(
                                    key="instrument_lists",
                                    title="Instrument Lists",
                                    grid_template_areas=[["instrument_lists"]],
                                    inline_key="instrument_lists",
                                    collapsible=False,
                                ),
                            ],
                        )
                    },
                ),
                # TODO Renable after adapting the code to use financial dataloader
                # Page(
                #     title="Financial Analysis",
                #     layouts={
                #         default(): Layout(
                #             grid_template_areas=[
                #                 ["sect_head_fin_summary", "sect_head_fin_summary"],
                #                 ["summary_table", "summary_table"],
                #                 ["sect_head_inc_cf_summary", "sect_head_inc_cf_summary"],
                #                 ["financials_graph", "cash_flow_analysis_chart"],
                #                 ["sect_head_profit_debt_summary", "sect_head_profit_debt_summary"],
                #                 ["profitability_ratios", "net_debt_and_ebitda_chart"],
                #                 ["sect_head_earnings", "sect_head_earnings"],
                #                 ["earnings_chart_ttm", "earnings_chart_ntm"],
                #             ],
                #             grid_template_rows=[
                #                 Style.px(40),
                #                 Style.px(800),
                #                 Style.px(40),
                #                 Style.px(500),
                #                 Style.px(40),
                #                 Style.px(500),
                #                 Style.px(40),
                #                 Style.px(500),
                #             ],
                #             inlines=[
                #                 Inline(key="summary_table", endpoint="summary_table"),
                #                 Inline(key="financials_graph", endpoint="financials_graph"),
                #                 Inline(key="profitability_ratios", endpoint="profitability_ratios"),
                #                 Inline(key="cash_flow_analysis_chart", endpoint="cash_flow_analysis_chart"),
                #                 Inline(key="net_debt_and_ebitda_chart", endpoint="net_debt_and_ebitda_chart"),
                #                 Inline(key="earnings_chart_ttm", endpoint="earnings_chart_ttm"),
                #                 Inline(key="earnings_chart_ntm", endpoint="earnings_chart_ntm"),
                #             ],
                #             sections=[
                #                 create_simple_section(key="sect_head_fin_summary", title="Financial Summary"),
                #                 create_simple_section(
                #                     key="sect_head_inc_cf_summary",
                #                     title="Income Statement and Cash Flow Summary Charts",
                #                 ),
                #                 create_simple_section(
                #                     key="sect_head_profit_debt_summary",
                #                     title="Profitability and Net Debt Summary Charts",
                #                 ),
                #                 create_simple_section(key="sect_head_earnings", title="Earnings Analysis"),
                #             ],
                #         )
                #     },
                # ),
                # Page(
                #     title="Valuation Ratios",
                #     layouts={
                #         default(): Layout(
                #             grid_template_areas=[
                #                 ["valuation_ratios_ranges", "valuation_ratios_related"],
                #             ],
                #             grid_template_columns=[Style.pct(50), Style.pct(50)],
                #             grid_template_rows=[Style.px(575), Style.fr(1)],
                #             inlines=[
                #                 Inline(key="valuation_ratios_ranges", endpoint="valuation_ratios_ranges"),
                #                 Inline(key="valuation_ratios_related", endpoint="valuation_ratios_related"),
                #             ],
                #         )
                #     },
                # ),
                Page(
                    title="Management",
                    layouts={
                        default(): Layout(
                            grid_template_areas=[
                                ["ticker", "refinitiv_identifier_code", "refinitiv_mnemonic_code", "identifier"],
                                ["base_color", "old_isins", "is_cash", "is_preferred_primary"],
                                ["tags", "tags", "related", "related"],
                            ],
                            grid_template_columns=[repeat(4, Style.fr(1))],
                            inlines=[
                                Inline(key="exchanges", endpoint="exchanges"),
                                Inline(key="related", endpoint="related_instruments"),
                            ],
                        )
                    },
                ),
                Page(
                    title=_("Hierarchy"),
                    layouts={
                        default(): Layout(
                            grid_template_areas=[["children"]],
                            inlines=[Inline(key="children", endpoint="children")],
                            grid_template_columns=[
                                "minmax(min-content, 1fr)",
                            ],
                            grid_template_rows=["1fr"],
                        ),
                    },
                ),
                Page(
                    title=_("News"),
                    layouts={
                        default(): Layout(
                            grid_template_areas=[["news"]],
                            inlines=[Inline(key="news", endpoint="news")],
                            grid_template_columns=[
                                "minmax(min-content, 1fr)",
                            ],
                            grid_template_rows=["1fr"],
                        ),
                    },
                ),
            ]
        )



# Generated by Django 5.0.8 on 2024-08-09 14:20

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("currency", "0001_initial"),
        ("geography", "0001_initial"),
        ("io", "0007_alter_exportsource_query_params"),
        ("tags", "0001_initial"),
        ("wbfdm", "0024_option_open_interest_10d_option_volume_10d_and_more"),
    ]

    operations = [
        migrations.AddField(
            model_name="instrument",
            name="is_primary",
            field=models.BooleanField(blank=True, null=True),
        ),
        migrations.AddConstraint(
            model_name="instrument",
            constraint=models.UniqueConstraint(
                condition=models.Q(("is_primary", True)),
                fields=("parent", "is_primary"),
                name="unique_instrument_primary",
            ),
        ),
    ]



# Generated by Django 4.2.7 on 2023-11-27 09:43

import colorfield.fields
import django.contrib.postgres.fields
import django.core.validators
import django.db.models.deletion
import django_better_admin_arrayfield.models.fields
import django_fsm
import mptt.fields
import wbcore.models.fields
from django.apps import apps
from django.db import migrations, models
from wbfdm.models.instruments.mixin.financials_computed import (
    InstrumentPriceComputedMixin,
)

state_operations = [
    migrations.CreateModel(
        name="Classification",
        fields=[
            (
                "id",
                models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID"),
            ),
            ("computed_str", models.CharField(blank=True, max_length=512, null=True, verbose_name="Name")),
            (
                "height",
                models.PositiveIntegerField(default=0, verbose_name="The height (leaf node have height 0)"),
            ),
            (
                "level_representation",
                models.CharField(max_length=256, verbose_name="Level Representation"),
            ),
            ("name", models.CharField(max_length=128, verbose_name="Name")),
            ("code_aggregated", models.CharField(max_length=64, verbose_name="Code Aggregated")),
            (
                "investable",
                models.BooleanField(default=True, help_text="Is this classification investable for us?"),
            ),
            (
                "description",
                models.TextField(
                    blank=True,
                    default="",
                    help_text="Give a basic definition and description",
                    verbose_name="Definition/Description",
                ),
            ),
            ("lft", models.PositiveIntegerField(editable=False)),
            ("rght", models.PositiveIntegerField(editable=False)),
            ("tree_id", models.PositiveIntegerField(db_index=True, editable=False)),
            ("level", models.PositiveIntegerField(editable=False)),
        ],
        options={
            "verbose_name": "Classification",
            "verbose_name_plural": "Classifications",
        },
    ),
    migrations.CreateModel(
        name="ClassificationGroup",
        fields=[
            (
                "id",
                models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID"),
            ),
            ("name", models.CharField(max_length=128, verbose_name="Name")),
            (
                "is_primary",
                models.BooleanField(
                    default=False,
                    help_text="Set to True if this classification must be used as default if not specified otherwise",
                    verbose_name="Primary",
                ),
            ),
            ("max_depth", models.IntegerField(default=0, verbose_name="Maximum Depth")),
            (
                "code_level_digits",
                models.IntegerField(default=2, verbose_name="The number of digits per code level"),
            ),
        ],
        options={
            "verbose_name": "Classification Group",
            "verbose_name_plural": "Classification Groups",
        },
    ),
    migrations.CreateModel(
        name="Deal",
        fields=[
            (
                "id",
                models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID"),
            ),
            (
                "type",
                models.CharField(
                    choices=[
                        ("DEAL", "Deal"),
                        ("FUNDING", "Funding"),
                        ("INVESTMENT", "Investment"),
                        ("PORTFOLIO_EXIT", "Portfolio Exit"),
                    ],
                    default="DEAL",
                    help_text="The deal type",
                    max_length=14,
                    verbose_name="Type",
                ),
            ),
            ("external_id", models.CharField(blank=True, max_length=64, null=True)),
            ("date", models.DateField()),
            ("transaction_amount", models.FloatField(help_text="Deal Size (in millions")),
            ("funding_round", models.CharField(max_length=128, verbose_name="Funding Round")),
            (
                "funding_round_category",
                models.CharField(max_length=128, verbose_name="Funding Round Category"),
            ),
            (
                "valuation",
                models.FloatField(
                    blank=True,
                    help_text="Valuation of the funded organization after this transaction (in Millions USD).",
                    null=True,
                    verbose_name="Valuaton",
                ),
            ),
            (
                "valuation_estimated",
                models.BooleanField(
                    default=False,
                    help_text="True if the valuation is an estimate",
                    verbose_name="Is valuation estimated",
                ),
            ),
            (
                "valuation_source_type",
                models.CharField(blank=True, help_text="The source type of the valuation", max_length=24, null=True),
            ),
            (
                "valuation_media_mention_source_urls",
                django.contrib.postgres.fields.ArrayField(
                    base_field=models.URLField(),
                    blank=True,
                    help_text="List of URLs used to source the valuation for the Media Mentions source type.",
                    null=True,
                    size=None,
                ),
            ),
        ],
        options={
            "abstract": False,
        },
        bases=(models.Model,),
    ),
    migrations.CreateModel(
        name="Exchange",
        fields=[
            (
                "id",
                models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID"),
            ),
            (
                "name",
                models.CharField(
                    blank=True,
                    help_text="Name of the Exchange.",
                    max_length=265,
                    null=True,
                    verbose_name="Exchange Name",
                ),
            ),
            (
                "opening_time",
                models.TimeField(
                    blank=True,
                    help_text="The opening time of the exchange",
                    null=True,
                    verbose_name="Exchange Opening time",
                ),
            ),
            (
                "closing_time",
                models.TimeField(
                    blank=True,
                    help_text="The closing time of the exchange",
                    null=True,
                    verbose_name="Exchange Closing time",
                ),
            ),
            (
                "mic_code",
                models.CharField(
                    blank=True,
                    help_text="Market Identifier Code.",
                    max_length=4,
                    null=True,
                    unique=True,
                    verbose_name="MIC (ISO)",
                ),
            ),
            (
                "mic_name",
                models.CharField(
                    blank=True,
                    help_text="Market Identifier Name.",
                    max_length=126,
                    null=True,
                    verbose_name="MIC (ISO) Name",
                ),
            ),
            (
                "operating_mic_code",
                models.CharField(
                    blank=True,
                    help_text="Operating Market Identifier Code.",
                    max_length=4,
                    null=True,
                    verbose_name="Operating MIC Code",
                ),
            ),
            (
                "operating_mic_name",
                models.CharField(
                    blank=True,
                    help_text="Operating Market Identifier Name.",
                    max_length=126,
                    null=True,
                    verbose_name="Operating MIC Name",
                ),
            ),
            (
                "bbg_exchange",
                models.CharField(
                    blank=True,
                    help_text="Bloomberg Exchange Code.",
                    max_length=4,
                    null=True,
                    verbose_name="BBG Exchange Code",
                ),
            ),
            (
                "bbg_composite_primary",
                models.BooleanField(
                    default=False,
                    help_text="Indicates the primary exchange for this BBG Composite Code.",
                    verbose_name="BBG Primary composite",
                ),
            ),
            (
                "bbg_composite",
                models.CharField(
                    blank=True,
                    help_text="Bloomberg Composite Code.",
                    max_length=4,
                    null=True,
                    verbose_name="BBG Composite Code",
                ),
            ),
            (
                "refinitiv_identifier_code",
                models.CharField(
                    blank=True,
                    help_text="Reuters Exchange Code.",
                    max_length=4,
                    null=True,
                    verbose_name="RIC Exchange Code",
                ),
            ),
            (
                "refinitiv_mnemonic",
                models.CharField(
                    blank=True,
                    help_text="Reuters Exchange Mnemonic Code.",
                    max_length=4,
                    null=True,
                    verbose_name="Refinitiv Mnemonic",
                ),
            ),
            (
                "website",
                models.URLField(
                    blank=True, help_text="The Website of the Exchange", null=True, verbose_name="Website"
                ),
            ),
            (
                "comments",
                models.TextField(
                    blank=True,
                    default="",
                    help_text="Any comments for this exchange",
                    verbose_name="Comments",
                ),
            ),
        ],
        options={
            "verbose_name": "Exchange",
            "verbose_name_plural": "Exchanges",
        },
    ),
    migrations.CreateModel(
        name="InstrumentClassificationRelatedInstrument",
        fields=[
            (
                "id",
                models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID"),
            ),
            (
                "related_instrument_type",
                models.CharField(
                    blank=True,
                    choices=[
                        ("PARTNER", "Partner"),
                        ("SUPPLIER", "Supplier"),
                        ("PEER", "Peer"),
                        ("COMPETITOR", "Competitor"),
                        ("BIGGEST_THREAT", "Biggest Threat"),
                        ("CUSTOMER", "Customer"),
                    ],
                    max_length=16,
                    null=True,
                ),
            ),
        ],
    ),
    migrations.CreateModel(
        name="InstrumentClassificationThroughModel",
        fields=[
            (
                "id",
                models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID"),
            ),
            ("tag_detail_endpoint", models.CharField(blank=True, max_length=255, null=True)),
            ("tag_representation", models.CharField(blank=True, max_length=512, null=True)),
            ("is_favorite", models.BooleanField(default=False)),
            ("reason", models.TextField(blank=True, default="", verbose_name="Reason for the choice")),
            (
                "pure_player",
                models.BooleanField(default=False, help_text="Pure Players Companies", verbose_name="Pure Player"),
            ),
            (
                "top_player",
                models.BooleanField(default=False, help_text="Top Players Companies", verbose_name="Top Player"),
            ),
            (
                "percent_of_revenue",
                models.DecimalField(
                    blank=True,
                    decimal_places=4,
                    max_digits=5,
                    null=True,
                    validators=[
                        django.core.validators.MinValueValidator(0),
                        django.core.validators.MaxValueValidator(1),
                    ],
                    verbose_name="% of revenue",
                ),
            ),
        ],
    ),
    migrations.CreateModel(
        name="InstrumentFavoriteGroup",
        fields=[
            (
                "id",
                models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID"),
            ),
            ("name", models.CharField(max_length=256)),
            (
                "public",
                models.BooleanField(
                    default=False, help_text="If set to True, this group will be available to everyone."
                ),
            ),
            (
                "primary",
                models.BooleanField(
                    default=False,
                    help_text="If set to True, this group will be set as default filter for instrument based viewset (Only one primary group allowed).",
                ),
            ),
        ],
        options={
            "abstract": False,
        },
    ),
    migrations.CreateModel(
        name="InstrumentList",
        fields=[
            (
                "id",
                models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID"),
            ),
            ("name", models.CharField(max_length=255)),
            ("identifier", models.CharField(blank=True, max_length=255, unique=True)),
            (
                "instrument_list_type",
                models.CharField(
                    blank=True,
                    choices=[("WATCH", "Watch List"), ("EXCLUSION", "Exclusion List")],
                    max_length=32,
                    null=True,
                ),
            ),
        ],
        options={
            "verbose_name": "Instrument List",
            "verbose_name_plural": "Instrument Lists",
            "permissions": (("administrate_instrumentlist", "Can administrate Instrument List"),),
        },
    ),
    migrations.CreateModel(
        name="InstrumentListThroughModel",
        fields=[
            (
                "id",
                models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID"),
            ),
            ("computed_str", models.CharField(blank=True, max_length=512, null=True, verbose_name="Name")),
            ("instrument_str", models.CharField(max_length=256)),
            ("from_date", models.DateField(blank=True, null=True)),
            ("to_date", models.DateField(blank=True, null=True)),
            ("comment", models.TextField(blank=True, default="")),
            ("validated", models.BooleanField(default=False)),
        ],
        options={
            "verbose_name": "Instrument in Instrument List",
            "notification_types": [
                (
                    "wbfdm.instrument_list_add",
                    "Instrument added to Instrument List",
                    "A notification when an instrument gets added to a list.",
                    True,
                    True,
                    True,
                )
            ],
        },
        bases=(models.Model,),
    ),
    migrations.CreateModel(
        name="InstrumentRequest",
        fields=[
            (
                "id",
                models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID"),
            ),
            (
                "instrument_type",
                models.CharField(
                    choices=[
                        ("ADR", "American Depository Receipt (ADR)"),
                        ("CF", "Close Ended Fund"),
                        ("EQ", "Equity"),
                        ("ETF", "Exchange Traded Fund"),
                        ("ETC", "Exchange Traded Commodity"),
                        ("ETN", "Exchange Traded Note"),
                        ("WARRANT", "Warrant"),
                        ("GDR", "Global Depository Receipt (GDR)"),
                        ("GS", "Genussschein"),
                        ("TRUST", "Investment Trust"),
                        ("NVDR", "Non-voting depository receipt (NVDR)"),
                        ("PREF", "Preference Share"),
                        ("COMP", "Company"),
                        ("QT", "Quote"),
                    ],
                    max_length=255,
                    verbose_name="Type",
                ),
            ),
            (
                "status",
                django_fsm.FSMField(
                    choices=[
                        ("PENDING", "Pending"),
                        ("APPROVED", "Approved"),
                        ("DENIED", "Denied"),
                        ("DRAFT", "Draft"),
                    ],
                    default="DRAFT",
                    help_text="The Request Status (default to Pending)",
                    max_length=50,
                    verbose_name="Status",
                ),
            ),
            ("notes", models.TextField(blank=True, null=True, verbose_name="Notes")),
            (
                "created",
                models.DateTimeField(auto_now_add=True, help_text="The request creation time", verbose_name="Created"),
            ),
            ("instrument_data", models.JSONField(default=dict, verbose_name="Instrument Data")),
        ],
        options={
            "abstract": False,
        },
    ),
    migrations.CreateModel(
        name="RelatedInstrumentThroughModel",
        fields=[
            (
                "id",
                models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID"),
            ),
            ("is_primary", models.BooleanField(default=False)),
            (
                "related_type",
                models.CharField(
                    choices=[
                        ("BENCHMARK", "Benchmark"),
                        ("PEER", "Peer"),
                        ("RISK_INSTRUMENT", "Risk Instrument"),
                    ],
                    default="BENCHMARK",
                    max_length=32,
                ),
            ),
        ],
    ),
    migrations.CreateModel(
        name="Cash",
        fields=[],
        options={
            "proxy": True,
            "indexes": [],
            "constraints": [],
        },
        bases=("wbfdm.instrument",),
    ),
    migrations.CreateModel(
        name="InstrumentPrice",
        fields=[
            (
                "id",
                models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID"),
            ),
            (
                "lock_all_dynamic_fields",
                models.BooleanField(default=False, help_text="If True, will not allow dynamic fields to be updated"),
            ),
            (
                "locked_dynamic_fields",
                django_better_admin_arrayfield.models.fields.ArrayField(
                    base_field=models.CharField(max_length=128), blank=True, default=list, size=None
                ),
            ),
            ("updated_at", models.DateTimeField(auto_now=True)),
            ("date", models.DateField(verbose_name="Date")),
            ("calculated", models.BooleanField(default=False, verbose_name="Is Calculated")),
            (
                "net_value",
                models.DecimalField(decimal_places=6, max_digits=16, verbose_name="Value (Net)"),
            ),
            (
                "gross_value",
                wbcore.models.fields.DynamicDecimalField(
                    blank=True, decimal_places=6, max_digits=16, null=True, verbose_name="Value (Gross)"
                ),
            ),
            (
                "outstanding_shares",
                wbcore.models.fields.DynamicDecimalField(
                    blank=True,
                    decimal_places=4,
                    help_text="The amount of outstanding share for this instrument",
                    max_digits=16,
                    null=True,
                    verbose_name="Outstanding Shares",
                ),
            ),
            (
                "outstanding_shares_consolidated",
                wbcore.models.fields.DynamicDecimalField(
                    blank=True,
                    decimal_places=4,
                    help_text="The amount of outstanding share for this instrument",
                    max_digits=16,
                    null=True,
                    verbose_name="Outstanding Shares (Consolidated)",
                ),
            ),
            (
                "volume",
                models.FloatField(
                    blank=True,
                    help_text="The Volume of the Asset on the price date of the Asset.",
                    null=True,
                    verbose_name="Volume",
                ),
            ),
            (
                "volume_50d",
                wbcore.models.fields.DynamicFloatField(
                    blank=True,
                    help_text="The Average Volume of the Asset over the last 50 days from the price date of the Asset.",
                    null=True,
                    verbose_name="Average Volume (50 Days)",
                ),
            ),
            (
                "volume_200d",
                wbcore.models.fields.DynamicFloatField(
                    blank=True,
                    help_text="The Average Volume of the Asset over the last 200 days from the price date of the Asset.",
                    null=True,
                    verbose_name="Average Volume (200 Days)",
                ),
            ),
            (
                "market_capitalization",
                models.FloatField(
                    blank=True,
                    help_text="The Market Capitalization of the Asset the price date of the Asset.",
                    null=True,
                    verbose_name="Market Capitalization",
                ),
            ),
            (
                "market_capitalization_consolidated",
                wbcore.models.fields.DynamicFloatField(
                    blank=True,
                    help_text="the consolidated market value of a company in local currency.",
                    null=True,
                    verbose_name="Market Capitalization (Consolidated)",
                ),
            ),
            (
                "lock_statistics",
                models.BooleanField(
                    default=False,
                    help_text="If True, a save will not override the beta, correlation and sharpe ratio",
                ),
            ),
            ("sharpe_ratio", models.FloatField(blank=True, null=True, verbose_name="Sharpe Ratio")),
            ("correlation", models.FloatField(blank=True, null=True, verbose_name="Correlation")),
            ("beta", models.FloatField(blank=True, null=True, verbose_name="Beta")),
            (
                "currency_fx_rate_to_usd",
                models.ForeignKey(
                    blank=True,
                    help_text="Rate to between instrument currency and USD",
                    null=True,
                    on_delete=django.db.models.deletion.PROTECT,
                    related_name="instrument_prices",
                    to="currency.currencyfxrates",
                    verbose_name="Instrument Currency Rate",
                ),
            ),
            (
                "import_source",
                models.ForeignKey(
                    blank=True,
                    null=True,
                    on_delete=django.db.models.deletion.SET_NULL,
                    to="io.importsource",
                ),
            ),
            (
                "instrument",
                models.ForeignKey(
                    blank=True,
                    null=True,
                    on_delete=django.db.models.deletion.PROTECT,
                    related_name="prices",
                    to="wbfdm.instrument",
                    verbose_name="Instrument",
                ),
            ),
        ],
        options={
            "verbose_name": "Instrument Price",
            "verbose_name_plural": "Instrument Prices",
            "indexes": [
                models.Index(fields=["calculated", "date", "instrument"], name="fdm_instrumentprice_base_idx"),
                models.Index(fields=["calculated", "instrument"], name="fdm_instrumentprice_idx1"),
                models.Index(fields=["instrument"], name="fdm_instrumentprice_idx2"),
            ],
        },
        bases=(
            InstrumentPriceComputedMixin,
            models.Model,
        ),
    ),
    migrations.AddField(
        model_name="instrument",
        name="additional_urls",
        field=django.contrib.postgres.fields.ArrayField(
            base_field=models.URLField(blank=True, null=True), blank=True, default=list, size=None
        ),
    ),
    migrations.AddField(
        model_name="instrument",
        name="alternative_names",
        field=django.contrib.postgres.fields.ArrayField(
            base_field=models.CharField(blank=True, max_length=255, null=True), default=list, size=None
        ),
    ),
    migrations.AddField(
        model_name="instrument",
        name="computed_str",
        field=models.CharField(blank=True, max_length=512, null=True, verbose_name="Name"),
    ),
    migrations.AddField(
        model_name="instrument",
        name="founded_year",
        field=models.IntegerField(blank=True, null=True, verbose_name="Founded Year"),
    ),
    migrations.AddField(
        model_name="instrument",
        name="headquarter_address",
        field=models.CharField(blank=True, help_text="The company Headquarter address", max_length=512, null=True),
    ),
    migrations.AddField(
        model_name="instrument",
        name="headquarter_city",
        field=models.ForeignKey(
            blank=True,
            help_text="The company's headquarter city",
            limit_choices_to={"level": 3},
            null=True,
            on_delete=django.db.models.deletion.SET_NULL,
            related_name="headquarters_of",
            to="geography.geography",
            verbose_name="Headquarter City",
        ),
    ),
    migrations.AddField(
        model_name="instrument",
        name="identifier",
        field=models.CharField(blank=True, max_length=255, null=True, verbose_name="Identifier"),
    ),
    migrations.AddField(
        model_name="instrument",
        name="import_source",
        field=models.ForeignKey(
            blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, to="io.importsource"
        ),
    ),
    migrations.AddField(
        model_name="instrument",
        name="is_cash",
        field=models.BooleanField(default=False),
    ),
    migrations.AddField(
        model_name="instrument",
        name="isin",
        field=models.CharField(
            blank=True,
            help_text="The ISIN provided by the bank.",
            max_length=12,
            null=True,
            unique=True,
            verbose_name="ISIN",
        ),
    ),
    migrations.AddField(
        model_name="instrument",
        name="last_update",
        field=models.DateTimeField(auto_now=True, null=True),
    ),
    migrations.AddField(
        model_name="instrument",
        name="name_repr",
        field=models.CharField(blank=True, max_length=255, null=True, verbose_name="Name (Representation)"),
    ),
    migrations.AddField(
        model_name="instrument",
        name="old_isins",
        field=django.contrib.postgres.fields.ArrayField(
            base_field=models.CharField(max_length=12),
            blank=True,
            default=list,
            help_text="These old ISINs are stored for this instrument to retrieve it more easily later.",
            size=None,
            verbose_name="Old ISINS",
        ),
    ),
    migrations.AddField(
        model_name="instrument",
        name="primary_url",
        field=models.URLField(blank=True, help_text="The Company website url", null=True),
    ),
    migrations.AddField(
        model_name="instrument",
        name="refinitiv_identifier_code",
        field=models.CharField(
            blank=True,
            help_text="Refinitiv Identifier Code",
            max_length=255,
            null=True,
            unique=True,
            verbose_name="RIC",
        ),
    ),
    migrations.AddField(
        model_name="instrument",
        name="refinitiv_mnemonic_code",
        field=models.CharField(
            blank=True,
            help_text="Refinitiv Datastream Mnemonic Code",
            max_length=255,
            null=True,
            unique=True,
            verbose_name="Refinitiv Datastream Mnemonic Code",
        ),
    ),
    migrations.AddField(
        model_name="instrument",
        name="refinitiv_ticker",
        field=models.CharField(
            blank=True,
            help_text="Refinitiv Refinitiv",
            max_length=255,
            null=True,
            unique=True,
            verbose_name="Refinitiv Ticker",
        ),
    ),
    migrations.AddField(
        model_name="instrument",
        name="sedol",
        field=models.CharField(
            blank=True,
            help_text="Stock Exchange Daily Official List",
            max_length=255,
            null=True,
            unique=True,
            verbose_name="SEDOL",
        ),
    ),
    migrations.AddField(
        model_name="instrument",
        name="tag_detail_endpoint",
        field=models.CharField(blank=True, max_length=255, null=True),
    ),
    migrations.AddField(
        model_name="instrument",
        name="tag_representation",
        field=models.CharField(blank=True, max_length=512, null=True),
    ),
    migrations.AddField(
        model_name="instrument",
        name="tags",
        field=models.ManyToManyField(blank=True, related_name="%(app_label)s_%(class)s_items", to="tags.tag"),
    ),
    migrations.AddField(
        model_name="instrument",
        name="ticker",
        field=models.CharField(
            blank=True,
            help_text="The Bloomberg ticker without the exchange (e.g. AAPL)",
            max_length=255,
            null=True,
            verbose_name="Ticker Bloomberg",
        ),
    ),
    migrations.AddField(
        model_name="instrument",
        name="valoren",
        field=models.CharField(
            blank=True,
            help_text="Valoren Number",
            max_length=255,
            null=True,
            unique=True,
            verbose_name="Valoren Number",
        ),
    ),
    migrations.AddField(
        model_name="relatedinstrumentthroughmodel",
        name="instrument",
        field=models.ForeignKey(
            on_delete=django.db.models.deletion.CASCADE,
            related_name="related_instruments_through",
            to="wbfdm.instrument",
        ),
    ),
    migrations.AddField(
        model_name="relatedinstrumentthroughmodel",
        name="related_instrument",
        field=models.ForeignKey(
            on_delete=django.db.models.deletion.CASCADE,
            related_name="dependent_instruments_through",
            to="wbfdm.instrument",
        ),
    ),
    migrations.AddField(
        model_name="instrumentrequest",
        name="created_instrument",
        field=models.OneToOneField(
            blank=True,
            null=True,
            on_delete=django.db.models.deletion.CASCADE,
            related_name="creation_request",
            to="wbfdm.instrument",
        ),
    ),
    migrations.AddField(
        model_name="instrumentrequest",
        name="handler",
        field=models.ForeignKey(
            blank=True,
            null=True,
            on_delete=django.db.models.deletion.SET_NULL,
            related_name="handled_instrument_requests",
            to="directory.person",
            verbose_name="Handler",
        ),
    ),
    migrations.AddField(
        model_name="instrumentrequest",
        name="requester",
        field=models.ForeignKey(
            blank=True,
            null=True,
            on_delete=django.db.models.deletion.SET_NULL,
            related_name="instrument_requests",
            to="directory.person",
            verbose_name="Requester",
        ),
    ),
    migrations.AddField(
        model_name="instrumentlistthroughmodel",
        name="import_source",
        field=models.ForeignKey(
            blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, to="io.importsource"
        ),
    ),
    migrations.AddField(
        model_name="instrumentlistthroughmodel",
        name="instrument",
        field=models.ForeignKey(
            blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, to="wbfdm.instrument"
        ),
    ),
    migrations.AddField(
        model_name="instrumentlistthroughmodel",
        name="instrument_list",
        field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to="wbfdm.instrumentlist"),
    ),
    migrations.AddField(
        model_name="instrumentfavoritegroup",
        name="instruments",
        field=models.ManyToManyField(
            blank=True,
            related_name="favorite_groups",
            to="wbfdm.instrument",
            verbose_name="Favorite Instruments Group",
        ),
    ),
    migrations.AddField(
        model_name="instrumentfavoritegroup",
        name="owner",
        field=models.ForeignKey(
            blank=True,
            null=True,
            on_delete=django.db.models.deletion.CASCADE,
            related_name="favorite_instruments_groups",
            to="directory.person",
        ),
    ),
    migrations.AddField(
        model_name="instrumentclassificationthroughmodel",
        name="classification",
        field=models.ForeignKey(
            on_delete=django.db.models.deletion.CASCADE,
            related_name="instruments_through",
            to="wbfdm.classification",
        ),
    ),
    migrations.AddField(
        model_name="instrumentclassificationthroughmodel",
        name="instrument",
        field=models.ForeignKey(
            on_delete=django.db.models.deletion.CASCADE,
            related_name="classifications_through",
            to="wbfdm.instrument",
        ),
    ),
    migrations.AddField(
        model_name="instrumentclassificationthroughmodel",
        name="related_instruments",
        field=models.ManyToManyField(
            blank=True, through="wbfdm.InstrumentClassificationRelatedInstrument", to="wbfdm.instrument"
        ),
    ),
    migrations.AddField(
        model_name="instrumentclassificationthroughmodel",
        name="tags",
        field=models.ManyToManyField(blank=True, related_name="%(app_label)s_%(class)s_items", to="tags.tag"),
    ),
    migrations.AddField(
        model_name="instrumentclassificationrelatedinstrument",
        name="classified_instrument",
        field=models.ForeignKey(
            on_delete=django.db.models.deletion.CASCADE,
            related_name="related_instruements",
            to="wbfdm.instrumentclassificationthroughmodel",
        ),
    ),
    migrations.AddField(
        model_name="instrumentclassificationrelatedinstrument",
        name="related_instrument",
        field=models.ForeignKey(
            on_delete=django.db.models.deletion.CASCADE,
            related_name="instrument_classification_related",
            to="wbfdm.instrument",
        ),
    ),
    migrations.AddField(
        model_name="exchange",
        name="city",
        field=models.ForeignKey(
            blank=True,
            help_text="The city where this Exchange is located at.",
            limit_choices_to={"level": 3},
            null=True,
            on_delete=django.db.models.deletion.PROTECT,
            related_name="portfolio_exchanges_city",
            to="geography.geography",
            verbose_name="City",
        ),
    ),
    migrations.AddField(
        model_name="exchange",
        name="country",
        field=models.ForeignKey(
            blank=True,
            limit_choices_to={"level": 1},
            null=True,
            on_delete=django.db.models.deletion.SET_NULL,
            to="geography.geography",
        ),
    ),
    migrations.AddField(
        model_name="deal",
        name="equity",
        field=models.ForeignKey(
            on_delete=django.db.models.deletion.CASCADE, related_name="deals", to="wbfdm.instrument"
        ),
    ),
    migrations.AddField(
        model_name="deal",
        name="import_source",
        field=models.ForeignKey(
            blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, to="io.importsource"
        ),
    ),
    migrations.AddField(
        model_name="deal",
        name="investors",
        field=models.ManyToManyField(
            blank=True,
            help_text="Investors",
            related_name="invested_deals",
            to="wbfdm.instrument",
            verbose_name="Investors",
        ),
    ),
    migrations.AddField(
        model_name="classification",
        name="group",
        field=models.ForeignKey(
            on_delete=django.db.models.deletion.CASCADE,
            related_name="classifications",
            to="wbfdm.classificationgroup",
            verbose_name="Classification Group",
        ),
    ),
    migrations.AddField(
        model_name="classification",
        name="parent",
        field=mptt.fields.TreeForeignKey(
            blank=True,
            null=True,
            on_delete=django.db.models.deletion.CASCADE,
            related_name="children",
            to="wbfdm.classification",
            verbose_name="Parent Classification",
        ),
    ),
    migrations.AddField(
        model_name="instrument",
        name="classifications",
        field=models.ManyToManyField(
            blank=True,
            limit_choices_to=models.Q(("level", models.F("group__max_depth"))),
            related_name="instruments",
            through="wbfdm.InstrumentClassificationThroughModel",
            to="wbfdm.classification",
            verbose_name="Classifications",
        ),
    ),
    migrations.AddField(
        model_name="instrument",
        name="exchange",
        field=models.ForeignKey(
            blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, to="wbfdm.exchange"
        ),
    ),
    migrations.AddField(
        model_name="instrument",
        name="related_instruments",
        field=models.ManyToManyField(
            blank=True,
            related_name="benchmarks_of",
            through="wbfdm.RelatedInstrumentThroughModel",
            to="wbfdm.instrument",
            verbose_name="The Related Instruments",
        ),
    ),
    migrations.AddField(
        model_name="instrument",
        name="base_color",
        field=colorfield.fields.ColorField(
            blank=True, default="#FF0000", image_field=None, max_length=64, null=True, samples=None
        ),
    ),
    migrations.AddField(
        model_name="instrument",
        name="issue_price",
        field=models.PositiveIntegerField(
            default=100,
            help_text="The initial issue price that is displayed on the factsheet",
            verbose_name="Issue Price",
        ),
    ),
    migrations.AddField(
        model_name="instrument",
        name="last_valuation_date",
        field=models.DateField(
            blank=True, help_text="Last Valuation Date", null=True, verbose_name="Last Valuation Date"
        ),
    ),
    migrations.AlterUniqueTogether(
        name="relatedinstrumentthroughmodel",
        unique_together={("instrument", "related_instrument", "is_primary", "related_type")},
    ),
    migrations.AddConstraint(
        model_name="instrumentclassificationthroughmodel",
        constraint=models.UniqueConstraint(
            fields=("instrument", "classification"), name="unique_classifiedinstruments"
        ),
    ),
    migrations.AddConstraint(
        model_name="instrumentprice",
        constraint=models.CheckConstraint(
            check=models.Q(("date__week_day__in", [1, 7]), _negated=True),
            name="wbfdm_instrumentprice_weekday_constraint",
        ),
    ),
    migrations.AddField(
        model_name="exchange",
        name="source",
        field=models.CharField(blank=True, max_length=64, null=True),
    ),
    migrations.AddField(
        model_name="exchange",
        name="source_id",
        field=models.CharField(blank=True, max_length=64, null=True),
    ),
    migrations.AlterField(
        model_name="instrumentrequest",
        name="instrument_type",
        field=models.CharField(
            choices=[
                ("ADR", "American Depository Receipt (ADR)"),
                ("CF", "Close Ended Fund"),
                ("EQ", "Equity"),
                ("ETF", "Exchange Traded Fund"),
                ("ETC", "Exchange Traded Commodity"),
                ("ETN", "Exchange Traded Note"),
                ("WARRANT", "Warrant"),
                ("GDR", "Global Depository Receipt (GDR)"),
                ("GS", "Genussschein"),
                ("TRUST", "Investment Trust"),
                ("NVDR", "Non-voting depository receipt (NVDR)"),
                ("PREF", "Preference Share"),
                ("COMP", "Company"),
                ("QT", "Quote"),
                ("Product", "Product"),
                ("Index", "Index"),
                ("ProductGroup", "Product Group"),
                ("Cash", "Cash"),
            ],
            max_length=255,
            verbose_name="Type",
        ),
    ),
    migrations.AddIndex(
        model_name="exchange",
        index=models.Index(fields=["source_id", "source"], name="wbfdm_excha_source__802ad7_idx"),
    ),
]


def get_operations():
    if apps.is_installed("wbportfolio"):
        return [
            migrations.RunSQL(
                sql='DROP TABLE "wbfdm_instrument" CASCADE',
                reverse_sql=migrations.RunSQL.noop,
            ),
            migrations.SeparateDatabaseAndState(state_operations=state_operations),
        ]
    return state_operations


def save_pms_instrument_id(apps, schema_editor):
    InstrumentLookup = apps.get_model("wbfdm", "InstrumentLookup")
    if apps.is_installed("wbportfolio"):
        for lp in InstrumentLookup.objects.filter(
            backend_path="wbfdm.contrib.dsws.backends.market_data.DSWSMarketDataBackend"
        ):
            try:
                lp.pms_instrument_id = lp.instrument.pms_instrument.id
                lp.save()
            except Exception:
                pass


class Migration(migrations.Migration):
    dependencies = [
        ("directory", "0006_employeremployeerelationship_position_name"),
        ("geography", "0001_initial"),
        ("tags", "0001_initial"),
        ("io", "0002_importsource_creator"),
        ("wbfdm", "0009_alter_controversy_flag_alter_controversy_initiated_and_more"),
    ]

    operations = [
        migrations.RemoveField(
            model_name="instrumentidentifier",
            name="instrument",
        ),
        migrations.AlterModelOptions(
            name="instrument",
            options={
                "permissions": (("administrate_instrument", "Can administrate Instrument"),),
                "verbose_name": "Instrument",
                "verbose_name_plural": "Instruments",
            },
        ),
        migrations.AddField(
            model_name="instrumentlookup",
            name="pms_instrument_id",
            field=models.IntegerField(blank=True, null=True),
        ),
        migrations.RunPython(save_pms_instrument_id),
        migrations.RemoveField(
            model_name="instrument",
            name="bbg_lookup",
        ),
        migrations.RemoveField(
            model_name="instrument",
            name="corporate_action_backend",
        ),
        migrations.RemoveField(
            model_name="instrument",
            name="esg_backend",
        ),
        migrations.RemoveField(
            model_name="instrument",
            name="financial_backend",
        ),
        migrations.RemoveField(
            model_name="instrument",
            name="isin_lookup",
        ),
        migrations.RemoveField(
            model_name="instrument",
            name="market_data_backend",
        ),
        migrations.RemoveField(
            model_name="instrument",
            name="officer_backend",
        ),
        migrations.RemoveField(
            model_name="instrument",
            name="ric_lookup",
        ),
        migrations.RemoveField(
            model_name="instrument",
            name="statement_backend",
        ),
        migrations.RemoveField(
            model_name="instrument",
            name="ticker_lookup",
        ),
        migrations.RemoveField(
            model_name="instrument",
            name="valoren_lookup",
        ),
        migrations.RemoveField(
            model_name="instrument",
            name="wkn_lookup",
        ),
        migrations.DeleteModel(
            name="InstrumentIdentifier",
        ),
        # Separate from state
        *get_operations(),
    ]



import typing

from wbcore.contrib.color.enums import WBColor
from wbcore.metadata.configs import display as dp
from wbcore.metadata.configs.display import DisplayViewConfig

if typing.TYPE_CHECKING:
    from wbfdm.viewsets import FinancialSummary


class FinancialSummaryDisplayViewConfig(DisplayViewConfig):
    view: "FinancialSummary"

    ESTIMATE_COLOR = "#D2E5F6"

    def get_list_display(self) -> dp.ListDisplay:
        def generate_formatting_rules(col_key: str) -> typing.Iterator[dp.FormattingRule]:
            yield dp.FormattingRule(
                condition=[("==", "eps_growth", "id")],
                style={
                    "borderBottom": "1px solid #000000",
                },
            )
            yield dp.FormattingRule(
                condition=[("==", "roic", "id")],
                style={
                    "borderBottom": "1px solid #000000",
                },
            )
            yield dp.FormattingRule(
                condition=[("==", "interest_coverage_ratio", "id")],
                style={
                    "borderBottom": "1px solid #000000",
                },
            )
            yield dp.FormattingRule(
                condition=[("==", "revenue_growth", "id")],
                style={
                    "color": "#9FA0A1",
                    "fontStyle": "italic",
                },
            )
            yield dp.FormattingRule(
                condition=[("==", "net_profit_growth", "id")],
                style={
                    "color": "#9FA0A1",
                    "fontStyle": "italic",
                },
            )
            yield dp.FormattingRule(
                condition=[("==", "eps_growth", "id")],
                style={
                    "color": "#9FA0A1",
                    "fontStyle": "italic",
                },
            )
            yield dp.FormattingRule(
                condition=[("==", "free_cash_flow_per_share_growth", "id")],
                style={
                    "color": "#9FA0A1",
                    "fontStyle": "italic",
                },
            )
            yield dp.FormattingRule(
                condition=[("==", "year", "id")],
                style={
                    "fontWeight": "bold",
                },
            )
            yield dp.FormattingRule(
                style={"color": WBColor.RED_DARK.value},
                condition=[("<", 0)],
            )
            if self.view.estimate_columns.get(col_key, False) is True:
                yield dp.FormattingRule(style={"background-color": self.ESTIMATE_COLOR})

        def generate_field(col: str) -> dp.Field:
            return dp.Field(
                key=col,
                label=col,
                width=80,
                formatting_rules=generate_formatting_rules(col),
                suppress_auto_size=False,
                resizable=False,
                movable=False,
                menu=False,
                size_to_fit=False,
            )

        return dp.ListDisplay(
            fields=[
                dp.Field(
                    key="label",
                    label=" ",
                    width=120,
                    suppress_auto_size=False,
                    resizable=False,
                    movable=False,
                    menu=False,
                    size_to_fit=False,
                    formatting_rules=[
                        dp.FormattingRule(
                            condition=[("==", "eps_growth", "id")],
                            style={
                                "borderBottom": "1px solid #000000",
                            },
                        ),
                        dp.FormattingRule(
                            condition=[("==", "roic", "id")],
                            style={
                                "borderBottom": "1px solid #000000",
                            },
                        ),
                        dp.FormattingRule(
                            condition=[("==", "interest_coverage_ratio", "id")],
                            style={
                                "borderBottom": "1px solid #000000",
                            },
                        ),
                        dp.FormattingRule(style={"background-color": "#ECECEC"}),
                        dp.FormattingRule(
                            condition=[("==", "year", "id")],
                            style={
                                "fontWeight": "bold",
                            },
                        ),
                    ],
                ),
                *map(generate_field, self.view.fiscal_columns),
            ],
            condensed=True,
            editable=False,
        )



from datetime import datetime
from typing import Any, Dict, Optional

from django.db import models
from wbcore.contrib.io.exceptions import DeserializationError
from wbcore.contrib.io.imports import ImportExportHandler

from .instrument import InstrumentImportHandler


class DealImportHandler(ImportExportHandler):
    MODEL_APP_LABEL: str = "wbfdm.Deal"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.instrument_handler = InstrumentImportHandler(self.import_source)

    def _deserialize(self, data):
        data["date"] = datetime.strptime(data["date"], "%Y-%m-%d").date()
        equity = self.instrument_handler.process_object(
            {**data["equity"], "instrument_type": "private_equity"}, read_only=True
        )[0]
        if not equity:
            raise DeserializationError("Private Equity couldn't be found with given data")
        data["equity"] = equity

        if investors := data.pop("investors", []):
            data["investors"] = [
                self.instrument_handler.process_object(investor_data, read_only=True)[0] for investor_data in investors
            ]

    def _get_instance(self, data: Dict[str, Any], history: Optional[models.QuerySet] = None, **kwargs) -> models.Model:
        if external_id := data.get("external_id", None):
            return self.model.objects.filter(external_id=external_id).first()
        qs = self.model.objects.filter(
            equity=data["equity"], date=data["date"], transaction_amount=data["transaction_amount"]
        )
        if qs.count() == 1:
            return qs.first()

    def _create_instance(self, data: Dict[str, Any], **kwargs) -> models.Model:
        investors = data.pop("investors", None)
        obj = self.model.objects.create(
            **data,
            import_source=self.import_source,
        )
        if investors:
            obj.investors.set([i for i in investors if i])
        return obj



from django.utils.translation import gettext as _
from wbcore.contrib.color.enums import WBColor
from wbcore.enums import Operator
from wbcore.metadata.configs import display as dp
from wbcore.metadata.configs.display import DisplayViewConfig

from wbfdm.enums import ESGControveryFlag


class InstrumentESGControversyDisplayViewConfig(DisplayViewConfig):
    def get_list_display(self) -> dp.ListDisplay:
        return dp.ListDisplay(
            fields=[
                dp.Field(key="initiated", label=_("Initiated"), width=100),
                dp.Field(key="review", label=_("Review"), width=100),
                dp.Field(key="headline", label=_("Headline")),
                dp.Field(key="narrative", label=_("Narrative")),
                dp.Field(key="source", label=_("Source")),
                dp.Field(key="status", label=_("Status")),
                dp.Field(key="type", label=_("Type")),
                dp.Field(key="assessment", label=_("Assessment")),
                dp.Field(key="response", label=_("Response")),
            ],
            formatting=[
                dp.Formatting(
                    formatting_rules=[
                        dp.FormattingRule(
                            style={"background-color": WBColor.GREEN_LIGHT.value},
                            condition=dp.Condition(operator=Operator.EQUAL, value=ESGControveryFlag.GREEN.value),
                        ),
                        dp.FormattingRule(
                            style={"background-color": WBColor.RED_LIGHT.value},
                            condition=dp.Condition(operator=Operator.EQUAL, value=ESGControveryFlag.RED.value),
                        ),
                        dp.FormattingRule(
                            style={"background-color": WBColor.YELLOW_LIGHT.value},
                            condition=dp.Condition(operator=Operator.EQUAL, value=ESGControveryFlag.YELLOW.value),
                        ),
                        dp.FormattingRule(
                            style={"background-color": WBColor.YELLOW_DARK.value},
                            condition=dp.Condition(operator=Operator.EQUAL, value=ESGControveryFlag.ORANGE.value),
                        ),
                        dp.FormattingRule(
                            style={"background-color": WBColor.GREEN_LIGHT.value},
                            condition=dp.Condition(operator=Operator.EQUAL, value=ESGControveryFlag.UNKNOWN.value),
                        ),
                    ],
                    column="flag",
                )
            ],
            legends=[
                dp.Legend(
                    label="Flag",
                    items=[
                        dp.LegendItem(icon="#99c140", label="Green"),
                        dp.LegendItem(icon="#fdfd96", label="Yellow"),
                        dp.LegendItem(icon="#db7b2b", label="Orange"),
                        dp.LegendItem(icon="#cc3232", label="Red"),
                        dp.LegendItem(icon="#d3d3d3", label="Unknown"),
                    ],
                )
            ],
        )


class InstrumentESGPAIDisplayViewConfig(DisplayViewConfig):
    def get_list_display(self) -> dp.ListDisplay:
        return dp.ListDisplay(
            fields=[
                dp.Field(key="section", label=_("Section"), width=90),
                dp.Field(key="asi", label=_("Adverse sustainability indicator")),
                dp.Field(key="metric", label=_("Metric")),
                dp.Field(key="factor", label=_("Factor"), width=450),
                dp.Field(key="value", label=_("Value"), width=120),
            ],
        )



from dataclasses import dataclass, field
from datetime import date
from typing import Any, Optional, Type

from django.db.models import Aggregate, FloatField, Sum


@dataclass
class MetricField:
    """
    A DTO Class to represent the metric subfield computed by metric backend for a given Metric Key.

    list_display_kwargs: Contains keyword argument to be inserted in a dp.Field
    """

    key: str
    label: str
    list_display_kwargs: dict[str, Any] = field(default_factory=dict)
    aggregate: Type[Aggregate] | None = Sum
    serializer_kwargs: dict[str, Any] = field(default_factory=dict)
    field_type: type = FloatField
    decorators: list[dict[str, str]] = field(
        default_factory=list
    )  # Set into a wbcore decorator accordingly (e.g. to show the metric into a specific currency)
    help_text: str | None = None  # define if the field needs to show a particular help text


@dataclass
class MetricKey:
    """
    a DTO class to represent the metric generated by a metric backend. It contains general information regarding the type of field the metric stored (e.g. Is percent? precision? etc..)
    """

    key: str
    label: str
    subfields: list[MetricField]
    extra_subfields: list[MetricField] = field(default_factory=list)
    additional_prefixes: list[str] = field(
        default_factory=list
    )  # Define the optional category/groups stored in the metric json (e.g if a variable "x" is stored as well as "past_x", a prefix would be "past")
    subfields_map: dict[str, MetricField] = field(init=False)
    subfields_filter_map: dict[str, str] = field(init=False)

    def __post_init__(self):
        subfields_map = dict()
        subfields_filter_map = dict()
        for subfield in self.subfields:
            subfields_map[f"{self.key}_{subfield.key}"] = subfield
            subfields_filter_map[f"{self.key}_{subfield.key}"] = subfield.key
            for prefix in self.additional_prefixes:
                subfields_map[f"{self.key}_{prefix}_{subfield.key}"] = subfield
                subfields_filter_map[f"{self.key}_{prefix}_{subfield.key}"] = f"{prefix}_{subfield.key}"
        object.__setattr__(self, "subfields_map", subfields_map)
        object.__setattr__(self, "subfields_filter_map", subfields_filter_map)

    def __hash__(self):
        return hash(self.key)

    def get_fields(self, with_prefixed_key: bool = False):
        """
        Returns a generator of key identifier and metric label containing in that metric

        Args:
            with_prefixed_key: Default to False. If True yield also the prefixed fields

        Returns:
            a generator of tuple of strings
        """
        for subfield in self.subfields:
            yield f"{self.key}_{subfield.key}", f"{self.label} {subfield.label}"
            if with_prefixed_key:
                for prefix in self.additional_prefixes:
                    yield f"{self.key}_{prefix}_{subfield.key}", f"{self.label} {subfield.label} ({prefix.title()})"


@dataclass
class Metric:
    """
    A DTO class to hold the metrics values computed by a metric backend. The object will be parsed by an orchestrator into a InstrumentMetric object
    """

    basket_id: int
    basket_content_type_id: int
    key: str
    metrics: dict
    date: Optional[date] = None
    instrument_id: int | None = None
    dependency_metrics: list["Metric"] = field(default_factory=list)



from datetime import date
from decimal import Decimal
from typing import Iterator

from django.db.models import Case, Value, When
from django.db.models.functions import Coalesce
from wbcore.contrib.currency.models import Currency, CurrencyFXRates
from wbcore.contrib.dataloader.dataloaders import Dataloader

from wbfdm.dataloaders.protocols import MarketDataProtocol
from wbfdm.dataloaders.types import MarketDataDict
from wbfdm.enums import Frequency, MarketData
from wbfdm.models.instruments.instrument_prices import InstrumentPrice

MarketDataMap = {
    "SHARES_OUTSTANDING": "internal_outstanding_shares",
    "OPEN": "net_value",
    "CLOSE": "net_value",
    "HIGH": "net_value",
    "LOW": "net_value",
    "BID": "net_value",
    "ASK": "net_value",
    "VOLUME": "internal_volume",
    "MARKET_CAPITALIZATION": "market_capitalization",
    "MARKET_CAPITALIZATION_CONSOLIDATED": "market_capitalization",
}

DEFAULT_VALUES = [MarketData[name] for name in MarketDataMap.keys()]


def _cast_decimal_to_float(value: float | Decimal) -> float:
    if isinstance(value, Decimal):
        value = float(value)
    return value


class MarketDataDataloader(MarketDataProtocol, Dataloader):
    def market_data(
        self,
        values: list[MarketData] | None = None,
        from_date: date | None = None,
        to_date: date | None = None,
        exact_date: date | None = None,
        frequency: Frequency = Frequency.DAILY,
        target_currency: str | None = None,
        apply_fx_rate: bool = True,
        **kwargs,
    ) -> Iterator[MarketDataDict]:
        """Get prices for instruments.

        Args:
            values (list[MarketData]): List of values to include in the results.
            from_date (date | None): The starting date for filtering prices. Defaults to None.
            to_date (date | None): The ending date for filtering prices. Defaults to None.
            frequency (Frequency): The frequency of the requested data

        Returns:
            Iterator[MarketDataDict]: An iterator of dictionaries conforming to the DailyValuationDict.
        """
        if not values:
            values = DEFAULT_VALUES
        values_map = {value.name: MarketDataMap[value.name] for value in values if value.name in MarketDataMap}
        calculated = kwargs.get("calculated", None)
        try:
            target_currency = Currency.objects.get(key=target_currency)
        except Currency.DoesNotExist:
            target_currency = None
        fx_rate = (
            Coalesce(
                CurrencyFXRates.get_fx_rates_subquery_for_two_currencies(
                    "date", "instrument__currency", target_currency
                ),
                Value(Decimal("1")),
            )
            if target_currency
            else Value(Decimal("1"))
        )

        prices = InstrumentPrice.objects.filter(instrument__in=self.entities)
        if calculated is not None:
            prices = prices.filter(calculated=calculated)
        else:
            prices = prices.filter_only_valid_prices()

        if exact_date:
            prices = prices.filter(date=exact_date)
        else:
            if from_date:
                prices = prices.filter(date__gte=from_date)
            if to_date:
                prices = prices.filter(date__lte=to_date)
        prices = prices.annotate_market_data().annotate(
            fx_rate=Case(When(calculated=False, then=fx_rate), default=None)
        )

        for row in prices.order_by("date").values(
            "date",
            "instrument",
            "calculated",
            "fx_rate",
            *set(values_map.values()),
        ):
            external_id = row.pop("instrument")
            val_date = row.pop("date")
            if row:
                fx_rate = row["fx_rate"]
                if apply_fx_rate and fx_rate is not None:
                    if row.get("net_value"):
                        row["net_value"] = row["net_value"] * fx_rate
                    if row.get("market_capitalization"):
                        row["market_capitalization"] = row["market_capitalization"] * float(fx_rate)
                    fx_rate = _cast_decimal_to_float(fx_rate)
                yield MarketDataDict(
                    id=f"{external_id}_{val_date}",
                    valuation_date=val_date,
                    instrument_id=external_id,
                    external_id=external_id,
                    source="wbfdm",
                    calculated=row["calculated"],
                    fx_rate=fx_rate,
                    **{MarketData[k].value: _cast_decimal_to_float(row[v]) for k, v in values_map.items()},
                )



from datetime import date

from oauthlib.uri_validate import query
from wbcore import filters as wb_filters

from wbfdm.models import InstrumentListThroughModel


class InstrumentListThroughModelFilterSet(wb_filters.FilterSet):
    validity_date = wb_filters.DateFilter(
        default=lambda: date.today(),
        help_text="The relationship validity date",
        label="validity_date",
        method="filter_validity_date",
        required=True,
    )

    validated = wb_filters.BooleanFilter(default=True, label="validated")

    def filter_validity_date(self, queryset, name, value):
        if value:
            return queryset.filter(timespan__contains=value)
        return query

    class Meta:
        model = InstrumentListThroughModel
        fields = {
            "instrument_list": ["exact"],
            "instrument": ["exact"],
            "validated": ["exact"],
        }



from wbcore.menus import ItemPermission, MenuItem

INSTRUMENT_LIST_MENUITEM = MenuItem(
    label="Instrument Lists",
    endpoint="wbfdm:instrumentlist-list",
    permission=ItemPermission(
        method=lambda request: request.user.is_internal, permissions=["wbfdm.view_instrumentlist"]
    ),
)



# Generated by Django 5.0.6 on 2024-07-08 11:12

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("wbfdm", "0023_instrument_unique_ric_instrument_unique_rmc_and_more"),
    ]

    operations = [
        migrations.AddField(
            model_name="option",
            name="open_interest_10d",
            field=models.FloatField(
                blank=True, help_text="Option Open Interest (10D)", null=True, verbose_name="Open Interest 10D"
            ),
        ),
        migrations.AddField(
            model_name="option",
            name="volume_10d",
            field=models.FloatField(blank=True, help_text="Option Volume (10D)", null=True, verbose_name="Volume 10D"),
        ),
        migrations.AddField(
            model_name="optionaggregate",
            name="open_interest_10d",
            field=models.FloatField(
                blank=True, help_text="Option Open Interest (10D)", null=True, verbose_name="Open Interest 10D"
            ),
        ),
        migrations.AddField(
            model_name="optionaggregate",
            name="volume_10d",
            field=models.FloatField(blank=True, help_text="Option Volume (10D)", null=True, verbose_name="Volume 10D"),
        ),
    ]



from wbcore.menus import ItemPermission, MenuItem

ClassificationMenuItem = MenuItem(
    label="Classification",
    endpoint="wbfdm:classification-list",
    permission=ItemPermission(
        permissions=["wbfdm.view_classification"],
        method=lambda request: request.user.is_internal,
    ),
)
ClassificationGroupMenuItem = MenuItem(
    label="Classification Group",
    endpoint="wbfdm:classificationgroup-list",
    permission=ItemPermission(
        permissions=["wbfdm.view_classificationgroup"],
        method=lambda request: request.user.is_internal,
    ),
)



from django.utils.translation import gettext as _
from wbcore.metadata.configs import display as dp
from wbcore.metadata.configs.display import DisplayViewConfig


class OfficerDisplayViewConfig(DisplayViewConfig):
    def get_list_display(self) -> dp.ListDisplay:
        return dp.ListDisplay(
            fields=[
                dp.Field(key="position", label=_("Position")),
                dp.Field(key="name", label=_("Name")),
                dp.Field(key="age", label=_("Age"), width=70),
                dp.Field(key="sex", label=_("Sex"), width=70),
                dp.Field(key="start", label=_("Start"), width=100),
            ],
        )



# Generated by Django 4.1.9 on 2023-07-19 07:42

import django.db.models.deletion
import wbfdm.models.fk_fields
from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("wbfdm", "0004_rename_financials_instrumentlookup_statements_and_more"),
    ]

    operations = [
        migrations.AddField(
            model_name="instrument",
            name="corporate_action_backend",
            field=wbfdm.models.fk_fields.CompositeForeignKey(
                null=True,
                null_if_equal=[],
                on_delete=django.db.models.deletion.DO_NOTHING,
                related_name="+",
                to="wbfdm.instrumentlookup",
                to_fields={
                    "corporate_actions": wbfdm.models.fk_fields.RawFieldValue(True),
                    "instrument_id": wbfdm.models.fk_fields.LocalFieldValue("id"),
                },
            ),
        ),
    ]



# Generated by Django 4.2.9 on 2024-03-29 14:17

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("wbfdm", "0014_alter_controversy_instrument"),
    ]

    operations = [
        migrations.AddIndex(
            model_name="instrument",
            index=models.Index(fields=["is_investable_universe"], name="instrument_investible_index"),
        ),
    ]



from jinja2 import Environment
from jinjasql import JinjaSql  # type: ignore


def get_environment(**options):
    # we except only SQL template, so we need to not escape special characters (possible XSS attack for HTML template)
    env = Environment(**options)  # noqa: S701
    return JinjaSql(env=env, param_style="format").env



from wbcore.menus import ItemPermission, MenuItem

INSTRUMENTFAVORITEGROUP_MENUITEM = MenuItem(
    label="Favorite Group",
    endpoint="wbfdm:favoritegroup-list",
    permission=ItemPermission(
        method=lambda request: request.user.is_internal, permissions=["wbfdm.view_instrumentfavoritegroup"]
    ),
)

CLASSIFIEDINSTRUMENTS_MENUITEM = MenuItem(
    label="Classified Instruments",
    endpoint="wbfdm:classifiedinstrument-list",
    permission=ItemPermission(method=lambda request: request.user.is_internal, permissions=["wbfdm.view_instrument"]),
)

INSTRUMENT_REQUEST_MENUITEM = MenuItem(
    label="Instrument Requests",
    endpoint="wbfdm:instrumentrequest-list",
    permission=ItemPermission(
        method=lambda request: request.user.is_internal, permissions=["wbfdm.view_instrumentrequest"]
    ),
    add=MenuItem(
        label="Create Instrument Request",
        endpoint="wbfdm:instrumentrequest-list",
        permission=ItemPermission(
            method=lambda request: request.user.is_internal, permissions=["wbfdm.add_instrumentrequest"]
        ),
    ),
)



# Generated by Django 5.0.12 on 2025-11-24 15:42

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('wbfdm', '0033_alter_controversy_review'),
    ]

    operations = [
        migrations.AlterField(
            model_name='instrumentlist',
            name='instrument_list_type',
            field=models.CharField(blank=True, choices=[('WATCH', 'Watch List'), ('EXCLUSION', 'Exclusion List'), ('INCLUSION', 'Inclusion List')], max_length=32, null=True),
        ),
    ]



import random
from datetime import date
from decimal import Decimal

import factory
from pandas.tseries.offsets import BDay
from wbcore.contrib.currency.models import CurrencyFXRates

from wbfdm.models import InstrumentPrice


def get_weekday(o):
    if (
        o.instrument.id
        and (prices := InstrumentPrice.objects.filter(instrument=o.instrument, calculated=o.calculated)).exists()
    ):
        latest_price = prices.latest("date").date
    else:
        latest_price = date(2020, 1, 1)
    return (latest_price + BDay(1)).date()


def get_coherent_net_value(o):
    previous_prices = InstrumentPrice.objects.filter(date__lt=o.date)
    if previous_prices.exists() and (last_price := previous_prices.latest("date")):
        return last_price.net_value * Decimal(
            1 + random.randrange(-5, 5)
        )  # Simulating a markov chain process of variance 10%
    return Decimal(100)


class InstrumentPriceFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = InstrumentPrice
        django_get_or_create = ("instrument", "date", "calculated")
        skip_postgeneration_save = True

    instrument = factory.SubFactory("wbfdm.factories.InstrumentFactory")
    date = factory.LazyAttribute(lambda o: get_weekday(o))
    currency_fx_rate_to_usd = factory.LazyAttribute(
        lambda o: CurrencyFXRates.objects.get_or_create(
            currency=o.instrument.currency, date=o.date, defaults={"value": Decimal(1.0)}
        )[0]
    )

    net_value = factory.Faker(
        "pydecimal", min_value=90, max_value=110, right_digits=6
    )  # we narrow down the range to simulate more realistic returns
    gross_value = factory.LazyAttribute(lambda o: o.net_value + Decimal(random.random()))

    calculated = False
    sharpe_ratio = factory.LazyAttribute(lambda o: random.random())
    correlation = factory.LazyAttribute(lambda o: random.random())
    beta = factory.LazyAttribute(lambda o: random.random())

    volume = factory.LazyAttribute(lambda o: random.randint(1000, 100000))
    volume_50d = factory.LazyAttribute(lambda o: random.randint(1000, 100000))
    volume_200d = factory.LazyAttribute(lambda o: random.randint(1000, 100000))
    market_capitalization = factory.LazyAttribute(lambda o: random.randint(10000, 100000))

    # custom_beta_180d = factory.Faker("pyfloat", positive=True, max_value=10e3)
    # custom_beta_1y = factory.Faker("pyfloat", positive=True, max_value=10e3)
    # custom_beta_2y = factory.Faker("pyfloat", positive=True, max_value=10e3)
    # custom_beta_3y = factory.Faker("pyfloat", positive=True, max_value=10e3)
    # custom_beta_5y = factory.Faker("pyfloat", positive=True, max_value=10e3)

    outstanding_shares = None
    volume_50d = factory.Faker("pyfloat", positive=True, max_value=10e3)
    volume_200d = factory.Faker("pyfloat", positive=True, max_value=10e3)

    # performance_1d = factory.Faker("pydecimal", min_value=-1, max_value=1)
    # performance_7d = factory.Faker("pydecimal", min_value=-1, max_value=1)
    # performance_30d = factory.Faker("pydecimal", min_value=-1, max_value=1)
    # performance_90d = factory.Faker("pydecimal", min_value=-1, max_value=1)
    # performance_365d = factory.Faker("pydecimal", min_value=-1, max_value=1)
    # performance_ytd = factory.Faker("pydecimal", min_value=-1, max_value=1)
    # performance_inception = factory.Faker("pydecimal", min_value=-1, max_value=1)

    @factory.post_generation
    def post_gen_currency_creation(self, create, extracted, **kwargs):
        CurrencyFXRates.objects.get_or_create(
            currency=self.instrument.currency, date=self.date, defaults={"value": Decimal(1.0)}
        )



from typing import Optional

from django.utils.translation import gettext as _
from wbcore.contrib.color.enums import WBColor
from wbcore.enums import Operator, Unit
from wbcore.metadata.configs import display as dp
from wbcore.metadata.configs.display.instance_display import (
    Display,
    Inline,
    Layout,
    Page,
    Section,
)
from wbcore.metadata.configs.display.instance_display.operators import default
from wbcore.metadata.configs.display.instance_display.shortcuts import (
    create_simple_display,
)

from wbfdm.models import InstrumentList


class InstrumentListDisplayConfig(dp.DisplayViewConfig):
    def get_list_display(self) -> Optional[dp.ListDisplay]:
        return dp.ListDisplay(
            fields=[
                dp.Field(key="name", label="Name"),
                dp.Field(key="instrument_list_type", label="Type"),
            ],
            legends=[
                dp.Legend(
                    items=[
                        dp.LegendItem(
                            icon=WBColor.RED_LIGHT.value,
                            label=_("Exclusion List"),
                        ),
                    ],
                )
            ],
            formatting=[
                dp.Formatting(
                    column="instrument_list_type",
                    formatting_rules=[
                        dp.FormattingRule(
                            style={"backgroundColor": WBColor.RED_LIGHT.value},
                            condition=dp.Condition(
                                operator=Operator.EQUAL, value=InstrumentList.InstrumentListType.EXCLUSION
                            ),
                        ),
                    ],
                ),
            ],
        )

    def get_instance_display(self) -> Display:
        return create_simple_display(
            grid_template_areas=[["name", "instrument_list_type"], ["instruments_section", "instruments_section"]],
            sections=[
                Section(
                    key="instruments_section",
                    title="Instruments",
                    collapsible=True,
                    collapsed=False,
                    display=Display(
                        pages=[
                            Page(
                                title="Instruments",
                                layouts={
                                    default(): Layout(
                                        grid_template_areas=[["instruments"]],
                                        inlines=[Inline(key="instruments", endpoint="instruments")],
                                    )
                                },
                            ),
                        ]
                    ),
                ),
            ],
        )


class InstrumentListThroughModelDisplayConfig(dp.DisplayViewConfig):
    def get_list_display(self) -> Optional[dp.ListDisplay]:
        fields = []
        if self.view.kwargs.get("instrument_id", None):
            fields = [dp.Field(key="instrument_list", label="Instrument List", width=Unit.PIXEL(250))]
        elif self.view.kwargs.get("instrument_list_id", None):
            fields = [
                dp.Field(key="instrument", label="Instrument", width=Unit.PIXEL(300)),
                dp.Field(key="instrument_str", label="Instrument Scraped", width=Unit.PIXEL(300)),
            ]
        fields.extend(
            [
                dp.Field(key="timespan", label="Timespan", width=Unit.PIXEL(250)),
                dp.Field(key="comment", label="Comment", width=Unit.PIXEL(350)),
            ]
        )
        return dp.ListDisplay(
            fields=fields,
            legends=[
                dp.Legend(
                    items=[
                        dp.LegendItem(
                            icon=WBColor.GREEN_LIGHT.value,
                            label=_("Validated"),
                        ),
                        dp.LegendItem(
                            icon=WBColor.YELLOW_DARK.value,
                            label=_("Not Validated"),
                        ),
                    ],
                )
            ],
            formatting=[
                dp.Formatting(
                    column="validated",
                    formatting_rules=[
                        dp.FormattingRule(
                            style={"backgroundColor": WBColor.GREEN_LIGHT.value},
                            condition=dp.Condition(operator=Operator.EQUAL, value=True),
                        ),
                        dp.FormattingRule(
                            style={"backgroundColor": WBColor.YELLOW_DARK.value},
                            condition=dp.Condition(operator=Operator.EQUAL, value=False),
                        ),
                    ],
                ),
            ],
        )

    def get_instance_display(self) -> Display:
        return create_simple_display(
            grid_template_areas=[
                ["instrument", "instrument_str"],
                ["instrument_list", "validated"],
                ["from_date", "to_date"],
                ["comment", "comment"],
            ]
        )



import factory

from wbfdm.models import InstrumentFavoriteGroup, RelatedInstrumentThroughModel


class RelatedInstrumentThroughModelFactory(factory.django.DjangoModelFactory):
    instrument = factory.SubFactory("wbfdm.factories.InstrumentFactory")
    related_instrument = factory.SubFactory("wbfdm.factories.InstrumentFactory")
    related_type = factory.Iterator(RelatedInstrumentThroughModel.RelatedTypeChoices.values)

    class Meta:
        model = RelatedInstrumentThroughModel
        django_get_or_create = ("instrument", "related_instrument", "related_type")


class InstrumentFavoriteGroupFactory(factory.django.DjangoModelFactory):
    name = factory.Faker("company")
    owner = factory.SubFactory("wbcore.contrib.directory.factories.PersonFactory")
    public = factory.Faker("pybool")
    primary = factory.Faker("pybool")

    class Meta:
        model = InstrumentFavoriteGroup
        skip_postgeneration_save = True

    @factory.post_generation
    def instruments(self, create, extracted, **kwargs):
        if not create:
            return

        if extracted:
            for instrument in extracted:
                self.instruments.add(instrument)



import logging
from contextlib import suppress
from datetime import date, timedelta
from typing import Callable

import pytz
from django.conf import settings
from django.core.exceptions import ObjectDoesNotExist
from django.db import IntegrityError, connections
from django.db.models import Max, QuerySet
from django.template.loader import get_template
from jinjasql import JinjaSql  # type: ignore
from mptt.exceptions import InvalidMove
from psycopg.errors import UniqueViolation
from tqdm import tqdm
from wbcore.contrib.currency.models import Currency
from wbcore.contrib.dataloader.utils import dictfetchall, dictfetchone
from wbcore.contrib.geography.models import Geography
from wbcore.utils.cache import mapping
from wbfdm.models.exchanges.exchanges import Exchange
from wbfdm.models.instruments.instruments import Instrument, InstrumentType

logger = logging.getLogger("pms")

BATCH_SIZE: int = 10000
instrument_type_map = {
    "ADR": "american_depository_receipt",
    "CF": "close_ended_fund",
    "EQ": "equity",
    "ET": "exchange_traded_fund",
    "ETC": "exchange_traded_commodity",
    "ETN": "exchange_traded_note",
    "EWT": "warrant",
    "GDR": "global_depository_receipt",
    "GNSH": "genussschein",
    "INVT": "investment_trust",
    "NVDR": "non_voting_depository_receipt",
    "PREF": "preference_share",
    "UT": "abc",
}


def get_dataloader_parameters(
    quote_code: str | None = None,
    exchange_code: str | None = None,
    rkd_code: str | None = None,
    ibes_code: str | None = None,
) -> dict:
    dataloader_parameters = dict()
    if quote_code:
        dataloader_parameters["adjustments"] = {
            "path": "wbfdm.contrib.qa.dataloaders.adjustments.DatastreamAdjustmentsDataloader",
            "parameters": quote_code,
        }
        dataloader_parameters["corporate_actions"] = {
            "path": "wbfdm.contrib.qa.dataloaders.corporate_actions.DatastreamCorporateActionsDataloader",
            "parameters": quote_code,
        }
        if exchange_code:
            dataloader_parameters["market_data"] = {
                "path": "wbfdm.contrib.qa.dataloaders.market_data.DatastreamMarketDataDataloader",
                "parameters": [
                    quote_code,
                    exchange_code,
                ],
            }

    if rkd_code:
        dataloader_parameters["officers"] = {
            "path": "wbfdm.contrib.qa.dataloaders.officers.RKDOfficersDataloader",
            "parameters": rkd_code,
        }
        dataloader_parameters["statements"] = {
            "path": "wbfdm.contrib.qa.dataloaders.statements.RKDStatementsDataloader",
            "parameters": rkd_code,
        }

    if ibes_code:
        dataloader_parameters["financials"] = {
            "path": "wbfdm.contrib.qa.dataloaders.financials.IBESFinancialsDataloader",
            "parameters": ibes_code,
        }
        dataloader_parameters["reporting_dates"] = {
            "path": "wbfdm.contrib.qa.dataloaders.reporting_dates.IbesReportingDateDataloader",
            "parameters": ibes_code,
        }

    return dataloader_parameters


def convert_data(data, parent_source: str | None = None) -> dict:
    if len(data.keys()) == 0:
        return data
    data["country_id"] = mapping(Geography.countries, "code_2").get(data["country_id"])
    data["currency_id"] = mapping(Currency.objects).get(data["currency_id"])
    if instrument_type_id := data.pop("instrument_type_id"):
        try:
            data["instrument_type_id"] = mapping(InstrumentType.objects)[
                instrument_type_map.get(instrument_type_id, instrument_type_id)
            ]
        except KeyError:
            data["instrument_type_id"] = InstrumentType.objects.get_or_create(
                key=instrument_type_id.lower(),
                defaults={"name": instrument_type_id.title(), "short_name": instrument_type_id.title()},
            )[0].pk
    data["exchange_id"] = mapping(Exchange.objects, "source_id", source="qa-ds2-exchange").get(
        str(data["exchange_code"])
    )
    data["dl_parameters"] = get_dataloader_parameters(
        data.pop("quote_code"), data.pop("exchange_code"), data.pop("rkd_code"), data.pop("ibes_code")
    )
    data["additional_urls"] = urls.split(",") if (urls := data.get("additional_urls")) else []
    if parent_source:
        data["parent_id"] = mapping(Instrument.objects, "source_id", source=parent_source).get(str(data["parent_id"]))
    if (is_primary := data.get("is_primary")) and isinstance(is_primary, str):
        data["is_primary"] = is_primary.lower().strip() == "y"
    data.pop("phone")
    return data


def get_instrument_from_data(data, parent_source: str | None = None) -> Instrument | None:
    if len(data.keys()) == 0:
        return None

    defaults = convert_data(data, parent_source)
    try:
        instrument = Instrument.objects.get(source=defaults["source"], source_id=defaults["source_id"])
        instrument.dl_parameters.update(
            defaults.pop("dl_parameters")
        )  # we need to update the existing dl parameters to not override its existing value
        for k, v in defaults.items():
            setattr(instrument, k, v)

    except Instrument.DoesNotExist:
        instrument = Instrument(**defaults)

    instrument.pre_save()
    instrument.computed_str = instrument.compute_str()
    if (
        instrument.name and instrument.currency and instrument.instrument_type
    ):  # we return an instrument only if it contains the basic name, currency and type
        return instrument


def _delist_existing_duplicates(instrument: Instrument) -> None:
    """Handle duplicate instruments by delisting existing entries"""
    unique_identifiers = ["refinitiv_identifier_code", "refinitiv_mnemonic_code", "isin", "sedol", "valoren", "cusip"]

    for identifier_field in unique_identifiers:
        if identifier := getattr(instrument, identifier_field):
            if instrument.delisted_date:  # if delisted, we unset the identifier that can lead to constraint error
                setattr(instrument, identifier_field, None)
            else:
                with suppress(Instrument.DoesNotExist):
                    duplicate = Instrument.objects.get(
                        is_security=True, delisted_date__isnull=True, **{identifier_field: identifier}
                    )
                    duplicate.delisted_date = date.today() - timedelta(days=1)
                    duplicate.save()


def _save_single_instrument(instrument: Instrument) -> None:
    """Attempt to save an instrument with duplicate handling"""
    try:
        instrument.save()
    except (UniqueViolation, IntegrityError) as e:
        if instrument.is_security:
            _delist_existing_duplicates(instrument)
            try:
                instrument.save()
                logger.info(f"{instrument} successfully saved after automatic delisting")
            except (UniqueViolation, IntegrityError) as e:
                logger.error(
                    "Instrument Synchronization: database integrity violation for security instrument.",
                    extra={"instrument": instrument, "detail": e},
                )
        else:
            logger.error(
                "Instrument Synchronization:  database integrity violation for non-security instrument.",
                extra={"instrument": instrument, "detail": e},
            )


def _bulk_create_instruments_chunk(instruments: list[Instrument], update_unique_identifiers: bool = False):
    update_fields = [
        "name",
        "dl_parameters",
        "description",
        "country",
        "currency",
        "parent",
        "primary_url",
        "additional_urls",
        "headquarter_address",
        "inception_date",
        "delisted_date",
        "exchange",
        "instrument_type",
        "computed_str",
        "search_vector",
        "trigram_search_vector",
        "is_primary",
    ]
    bulk_update_kwargs = {"update_conflicts": True}
    if update_unique_identifiers:
        update_fields.extend(
            [
                "refinitiv_identifier_code",
                "refinitiv_mnemonic_code",
                "isin",
                "sedol",
                "valoren",
                "cusip",
            ]
        )
        bulk_update_kwargs = {"ignore_conflicts": True}
    try:
        Instrument.objects.bulk_create(
            instruments, update_fields=update_fields, unique_fields=["source", "source_id"], **bulk_update_kwargs
        )
    except IntegrityError:
        # we caught an integrity error on the bulk save, so we try to save one by one
        logger.error(
            "we detected an integrity error while bulk saving instruments. We save them one by one and delist the already existing instrument from the db if we can. "
        )
        for instrument in instruments:
            _save_single_instrument(instrument)


def update_instruments(sql_name: str, parent_source: str | None = None, context=None, debug: bool = False, **kwargs):
    template = get_template(sql_name, using="jinja").template  # type: ignore
    if context is None:
        context = {}
    query, params = JinjaSql(param_style="format").prepare_query(template, context)
    instruments = []

    gen = dictfetchall(connections["qa"].cursor().execute(query, params))
    if debug:
        gen = tqdm(gen)
    # we update in batch to be sure to not exhaust resources
    for row in gen:
        with suppress(TypeError):  # we don't fail if the given data doesn't satisfy the schema
            if instrument := get_instrument_from_data(row, parent_source=parent_source):
                instruments.append(instrument)

        if len(instruments) >= BATCH_SIZE:
            _bulk_create_instruments_chunk(instruments, **kwargs)
            instruments = []

    _bulk_create_instruments_chunk(instruments, **kwargs)


def update_or_create_item(
    external_id: int, get_item: Callable, source: str, parent_source: str | None = None
) -> Instrument | None:
    defaults = convert_data(get_item(external_id), parent_source=parent_source)

    dl_parameters = defaults.pop("dl_parameters", {})
    defaults.pop("source", None)
    defaults.pop("source_id", None)
    try:
        instrument, _ = Instrument.objects.update_or_create(
            source=source,
            source_id=external_id,
            defaults=defaults,
        )
        instrument.dl_parameters.update(dl_parameters)
        _save_single_instrument(instrument)
        return instrument

    except (
        InvalidMove,
        ObjectDoesNotExist,
    ):  # we might encounter a object does not exist error in case the inserted parent does not exist yet in our db which might happen if it was just deleted because invalid
        logger.warning(
            f"We encountered a MPTT ill-formed node for source ID {external_id}. Rebuilding the tree might be necessary."
        )


def get_item(external_id: int, template_name: str) -> dict:
    template = get_template(template_name, using="jinja").template  # type: ignore
    query, params = JinjaSql(param_style="format").prepare_query(template, {"source_id": external_id})
    return dictfetchone(connections["qa"].cursor().execute(query, params))


def trigger_partial_update(
    queryset: QuerySet,
    last_update_field: str,
    id_field: str,
    table_change_name: str,
    update: Callable,
    update_or_create_item: Callable,
):
    max_last_updated = queryset.aggregate(max_last_updated=Max(last_update_field)).get("max_last_updated")

    if max_last_updated is None:
        update()

    else:
        with connections["qa"].cursor() as cursor:
            cursor.execute(
                "SELECT MAX(last_user_update) FROM sys.dm_db_index_usage_stats WHERE OBJECT_NAME(object_id) = %s",
                (table_change_name,),
            )
            max_last_updated_qa = (
                pytz.timezone(settings.TIME_ZONE).localize(result[0]) if (result := cursor.fetchone()) else None
            )
            if max_last_updated_qa and max_last_updated_qa > max_last_updated:
                for _, security_id in cursor.execute(
                    f"SELECT UpdateFlag_, {id_field} FROM {table_change_name}"  # noqa: S608
                ).fetchall():
                    try:
                        update_or_create_item(security_id)
                    except Exception as e:
                        logger.error(
                            "Instrument Partial Update: Error updating instrument.",
                            extra={"security_id": security_id, "detail": e},
                        )



# Generated by Django 5.2.9 on 2026-02-03 15:39

import django.contrib.postgres.constraints
import django.contrib.postgres.fields.ranges
from django.db import migrations
from psycopg.types.range import DateRange

from wbfdm.models import InstrumentListThroughModel
import django.contrib.postgres.constraints
from django.db import migrations, models

def migrate_timespan(apps, schema_editor):
    InstrumentListThroughModel = apps.get_model('wbfdm.InstrumentListThroughModel')
    objs = []
    for rel in InstrumentListThroughModel.objects.all():
        rel.timespan = DateRange(rel.from_date, rel.to_date)
        objs.append(rel)

    InstrumentListThroughModel.objects.bulk_update(objs, ["timespan"])
class Migration(migrations.Migration):

    dependencies = [
        ('io', '0009_alter_importedobjectproviderrelationship_unique_together_and_more'),
        ('wbfdm', '0037_relatedinstrumentthroughmodel_unique_instrument_relationship_and_more'),
        ('currency', '0002_alter_currencyfxrates_unique_together_and_more'),
        ('geography', '0002_geography_geography_geography_tree_i739a'),
        ('tags', '0002_alter_tag_unique_together_tag_unique_tag'),
    ]

    operations = [

        migrations.AddField(
            model_name='instrumentlistthroughmodel',
            name='timespan',
            field=django.contrib.postgres.fields.ranges.DateRangeField(default=None, verbose_name='Timespan', null=True, blank=True),
            preserve_default=False,
        ),
        migrations.RunPython(migrate_timespan),
        migrations.RemoveField(
            model_name='instrumentlistthroughmodel',
            name='from_date',
        ),
        migrations.RemoveField(
            model_name='instrumentlistthroughmodel',
            name='to_date',
        ),
        migrations.AddConstraint(
            model_name='instrumentlistthroughmodel',
            constraint=django.contrib.postgres.constraints.ExclusionConstraint(expressions=[('timespan', '&&'), ('instrument', '='), ('instrument_list', '=')], name='exclude_overlapping_instrument_list'),
        ),
        migrations.AlterField(
            model_name='instrumentlistthroughmodel',
            name='timespan',
            field=django.contrib.postgres.fields.ranges.DateRangeField(default=None, verbose_name='Timespan'),
            preserve_default=False,
        ),
        migrations.RemoveConstraint(
            model_name='instrumentlistthroughmodel',
            name='unique_instrument_per_list',
        ),
        migrations.RemoveConstraint(
            model_name='instrumentlistthroughmodel',
            name='exclude_overlapping_instrument_list',
        ),
        migrations.AddField(
            model_name='instrumentlist',
            name='constituent_loader_path',
            field=models.CharField(blank=True,
                                   help_text='If set, load the Instrument queryset returned by this function and set them as the new instrument list constituent',
                                   max_length=512, null=True),
        ),
        migrations.AddConstraint(
            model_name='instrumentlistthroughmodel',
            constraint=django.contrib.postgres.constraints.ExclusionConstraint(
                condition=models.Q(('instrument__isnull', False)),
                expressions=[('timespan', '&&'), ('instrument', '='), ('instrument_list', '=')],
                name='exclude_overlapping_instrument_list'),
        ),
        migrations.AddField(
            model_name='instrumentlist',
            name='constituent_loader_kwargs',
            field=models.JSONField(default=dict, blank=True),
        ),
        migrations.AddField(
            model_name='instrument',
            name='is_preferred_primary',
            field=models.BooleanField(default=False),
        ),
        migrations.AddConstraint(
            model_name='instrument',
            constraint=models.UniqueConstraint(condition=models.Q(('is_preferred_primary', True)),
                                               fields=('parent', 'is_preferred_primary'),
                                               name='unique_instrument_preferred_primary'),
        ),
    ]



from dataclasses import dataclass
from datetime import date


@dataclass
class PriceDTO:
    pk: str
    instrument: int
    date: date
    open: float
    close: float
    high: float
    low: float
    volume: float
    market_capitalization: float
    outstanding_shares: float

    @property
    def net_value(self) -> float:
        return self.close


@dataclass
class AdjustmentDTO:
    pk: str
    date: date
    adjustment_factor: float
    cumulative_adjustment_factor: float


@dataclass
class DividendDTO:
    pk: str
    rate: float
    ex_dividend_date: date
    payment_date: date



from dataclasses import dataclass
from datetime import date

import numpy as np
import pandas as pd
from django.utils.functional import cached_property
from wbcore.contrib.currency.models import Currency, CurrencyFXRates

from wbfdm.enums import ESG
from wbfdm.models import Instrument

from .enums import AggregationMethod, ESGAggregation
from .utils import get_esg_df


@dataclass
class Log:
    series: pd.Series
    label: str
    is_percent: bool = False
    precision: int = 4
    group: str | None = None


class DataLoader:
    def __init__(
        self, weights: pd.Series, esg_data: pd.Series, val_date: date, total_value_fx_usd: pd.Series | None = None
    ):
        self.weights = weights
        self.val_date = val_date
        try:
            self.fx_rate_usd_to_eur = float(
                CurrencyFXRates.objects.get(currency=Currency.objects.get(key="EUR"), date=val_date).value
            )
        except (CurrencyFXRates.DoesNotExist, Currency.DoesNotExist):
            self.fx_rate_usd_to_eur = 1.0
        self.total_value_fx_usd = total_value_fx_usd
        self.instruments = Instrument.objects.filter(id__in=self.weights.index)
        self.empty_series = pd.Series(1.0, index=weights.index, dtype="float64")
        self.esg_data = esg_data
        self.weights_in_coverage = (
            self.weights.loc[~esg_data.reindex(self.weights.index, fill_value=None).isnull()]
        ).rename("weights_in_coverage")
        self.intermediary_logs: list[Log] = []
        self.extra_esg_data_logs: list[Log] = []

    @cached_property
    def enterprise_value_included_cash(self) -> pd.Series:
        data = get_esg_df(self.instruments, ESG.EVIC_EUR) * 1000000
        self.extra_esg_data_logs.append(Log(series=data.rename("enterprise_value_included_cash"), label="EVIC (EUR)"))
        return data

    @cached_property
    def current_value_investment_factor(self) -> pd.Series:
        data = get_esg_df(self.instruments, ESG.CVI_FACTOR)
        data = data.reindex(self.weights.index, fill_value=1.0)
        self.extra_esg_data_logs.append(Log(series=data.rename("current_value_investment_factor"), label="CVI Factor"))
        return data

    @cached_property
    def nace_section_code(self) -> pd.Series:
        data = get_esg_df(self.instruments, ESG.NACE_SECTION_CODE)
        self.extra_esg_data_logs.append(Log(series=data.rename("nace_section_code"), label="NACE_SECTION_CODE"))
        return data

    def _get_percentage_sum(self, mask_value: str) -> pd.Series:
        df = self.weights_in_coverage.copy()
        df.loc[self.esg_data != mask_value] = 0.0
        return df.dropna()

    def _get_weighted_avg_normalized(self) -> pd.Series:
        weights_normalized = self.weights_in_coverage / self.weights_in_coverage.sum()
        self.intermediary_logs.append(
            Log(
                series=weights_normalized.rename("weights_normalized"),
                label="Rebased Weights",
                is_percent=True,
                precision=2,
            )
        )
        return (weights_normalized * self.esg_data).dropna()

    def _get_weighted_avg_normalized_per_category(self) -> pd.Series:
        df = pd.concat([self.weights, self.nace_section_code], keys=["weighting", "nace_section_code"], axis=1)
        sector_codes = ["C", "D", "F", "G", "J", "K", "M", "N", "Q", "S"]
        res = pd.Series(np.nan, index=df.index)
        rebased_weights_logs_per_sector = []
        weighted_average_logs_per_sector = []
        for sector_code in sector_codes:
            try:
                dff = df[df["nace_section_code"] == sector_code]
                weights_normalized_per_sector = dff["weighting"] / dff["weighting"].sum()
            except (TypeError, KeyError):
                weights_normalized_per_sector = pd.Series(np.nan, index=self.weights.index)

            rebased_weights_logs_per_sector.append(
                Log(
                    series=weights_normalized_per_sector.rename(f"weights_normalized_sector_{sector_code.lower()}"),
                    label=sector_code,
                    group="Weight rebased per NACE sector",
                    is_percent=True,
                    precision=2,
                )
            )
            weighted_average_per_sector = weights_normalized_per_sector * self.esg_data
            weighted_average_logs_per_sector.append(
                Log(
                    series=weighted_average_per_sector.rename(f"weighted_average_sector_{sector_code.lower()}"),
                    label=sector_code,
                    group="Weight Average per NACE sector",
                    precision=4,
                )
            )

            res[weighted_average_per_sector.dropna().index] = weighted_average_per_sector.dropna()
        self.intermediary_logs.extend(rebased_weights_logs_per_sector)
        self.intermediary_logs.extend(weighted_average_logs_per_sector)
        return res.dropna()

    def _get_investor_allocation(self, exposure: pd.Series) -> pd.Series:
        exposure_eur = exposure * self.fx_rate_usd_to_eur
        self.intermediary_logs.append(Log(series=exposure.rename("exposure"), label="Exposure"))
        self.intermediary_logs.append(Log(series=exposure_eur.rename("exposure_eur"), label="Exposure (EUR)"))

        evic_eur = self.enterprise_value_included_cash
        cvi_factor = self.current_value_investment_factor

        exposure_with_cvi = exposure_eur * cvi_factor
        self.intermediary_logs.append(
            Log(series=exposure_with_cvi.rename("exposure_with_cvi"), label="Exposure With CVI")
        )
        try:
            rebase_factor = exposure_with_cvi.loc[self.weights_in_coverage.index].sum() / exposure_with_cvi.sum()
            exposure_with_cvi_normalized = exposure_with_cvi.loc[self.weights_in_coverage.index] / rebase_factor
        except ZeroDivisionError:
            exposure_with_cvi_normalized = exposure_with_cvi

        self.intermediary_logs.append(
            Log(
                series=exposure_with_cvi_normalized.rename("exposure_with_cvi_normalized"),
                label="Exposure With CVI (Normalized)",
            )
        )

        attribution_factor = exposure_with_cvi_normalized / evic_eur
        self.intermediary_logs.append(
            Log(series=attribution_factor.rename("attribution_factor"), label="Attribution Factor", is_percent=True)
        )

        return (self.esg_data * attribution_factor).dropna()

    def _get_investor_allocation_per_million(self, exposure: pd.Series) -> pd.Series:
        self.intermediary_logs.append(Log(series=exposure.rename("exposure"), label="Exposure"))

        evic_eur = self.enterprise_value_included_cash
        cvi_factor = self.current_value_investment_factor

        exposure_with_cvi = exposure * cvi_factor
        self.intermediary_logs.append(
            Log(series=exposure_with_cvi.rename("exposure_with_cvi"), label="Exposure With CVI")
        )
        try:
            rebase_factor = exposure_with_cvi.loc[self.weights_in_coverage.index].sum() / exposure_with_cvi.sum()
            exposure_with_cvi_normalized = exposure_with_cvi.loc[self.weights_in_coverage.index] / rebase_factor
        except ZeroDivisionError:
            exposure_with_cvi_normalized = exposure_with_cvi

        self.intermediary_logs.append(
            Log(
                series=exposure_with_cvi_normalized.rename("exposure_with_cvi_normalized"),
                label="Exposure With CVI (Normalized)",
            )
        )

        weights_normalized = exposure_with_cvi_normalized / exposure_with_cvi_normalized.sum()
        self.intermediary_logs.append(
            Log(
                series=weights_normalized.rename("weights_normalized"),
                label="Weight based on CVI",
                is_percent=True,
                precision=2,
            )
        )
        cvi_based_exposure_per_million = weights_normalized * 1e6
        self.intermediary_logs.append(
            Log(
                series=cvi_based_exposure_per_million.rename("cvi_based_exposure_per_million"),
                label="CVI-based exposure for EUR 1mn",
            )
        )

        attribution_factor = cvi_based_exposure_per_million / evic_eur
        self.intermediary_logs.append(
            Log(series=attribution_factor.rename("attribution_factor"), label="Attribution Factor", is_percent=True)
        )

        return (self.esg_data * attribution_factor).dropna()

    def compute(self, esg_aggregation: ESGAggregation):
        aggregation_method = esg_aggregation.get_aggregation()
        if aggregation_method == AggregationMethod.PERCENTAGE_SUM:
            # we need to apply a mask
            if esg_aggregation in [ESGAggregation.LACK_OF_PROCESS_AND_COMPLIANCE_OF_UN_PRINCIPLES]:
                mask = "No Evidence"
            else:
                mask = "Yes"

            return self._get_percentage_sum(mask)
        if aggregation_method == AggregationMethod.WEIGHTED_AVG_NORMALIZED:
            return self._get_weighted_avg_normalized()
        if aggregation_method == AggregationMethod.WEIGHTED_AVG_CATEGORY_NORMALIZED:
            return self._get_weighted_avg_normalized_per_category()
        if aggregation_method == AggregationMethod.INVESTOR_ALLOCATION:
            return self._get_investor_allocation(self.total_value_fx_usd)
        if aggregation_method == AggregationMethod.INVESTOR_ALLOCATION_PER_MILLION:
            return self._get_investor_allocation_per_million(self.weights * 1000000)
        else:
            raise ValueError("Aggregation Method not supported")



# Generated by Django 5.2.9 on 2025-12-16 15:15

from django.db import migrations, models



def cleanup_duplicates(apps, schema_editor):
    RelatedInstrumentThroughModel = apps.get_model('wbfdm.RelatedInstrumentThroughModel')
    for row in RelatedInstrumentThroughModel.objects.values("instrument", "related_instrument", "related_type").annotate(c=models.Count("*")).filter(c__gt=1):
        qs = RelatedInstrumentThroughModel.objects.filter(
            instrument=row["instrument"],
            related_instrument=row["related_instrument"],
            related_type=row["related_type"]
        )
        qs.exclude(id=qs.first().id).delete()


class Migration(migrations.Migration):

    dependencies = [
        ('wbfdm', '0036_alter_relatedinstrumentthroughmodel_unique_together_and_more'),
    ]

    operations = [
        migrations.RunPython(cleanup_duplicates),
        migrations.AddConstraint(
            model_name='relatedinstrumentthroughmodel',
            constraint=models.UniqueConstraint(fields=('instrument', 'related_instrument', 'related_type'), name='unique_instrument_relationship'),
        ),
        migrations.AddConstraint(
            model_name='relatedinstrumentthroughmodel',
            constraint=models.UniqueConstraint(condition=models.Q(('is_primary', True)), fields=('instrument', 'related_type'), name='unique_primary_instrument_relationship'),
        ),
    ]



# Generated by Django 5.2.9 on 2025-12-09 08:53

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('currency', '0001_initial'),
        ('geography', '0002_geography_geography_geography_tree_i739a'),
        ('io', '0008_importsource_resource_kwargs'),
        ('tags', '0001_initial'),
        ('wbfdm', '0034_alter_instrumentlist_instrument_list_type'),
    ]

    operations = [
        migrations.AlterField(
            model_name='instrument',
            name='related_instruments',
            field=models.ManyToManyField(blank=True, related_name='benchmarks_of', through='wbfdm.RelatedInstrumentThroughModel', through_fields=('instrument', 'related_instrument'), to='wbfdm.instrument', verbose_name='The Related Instruments'),
        ),
        migrations.AlterField(
            model_name='instrumentclassificationthroughmodel',
            name='related_instruments',
            field=models.ManyToManyField(blank=True, limit_choices_to=models.Q(('instrument_type__is_classifiable', True), ('level', 0)), through='wbfdm.InstrumentClassificationRelatedInstrument', through_fields=('classified_instrument', 'related_instrument'), to='wbfdm.instrument'),
        ),
        migrations.AddIndex(
            model_name='instrument',
            index=models.Index(fields=['tree_id', 'lft'], name='wbfdm_instrument_tree_id_lf91c'),
        ),
    ]



import logging
from contextlib import suppress
from datetime import date, timedelta
from decimal import Decimal

import numpy as np
import pandas as pd
from django.core.exceptions import MultipleObjectsReturned, ValidationError
from django.core.validators import DecimalValidator
from django.db.models import (
    AutoField,
    Case,
    Exists,
    ExpressionWrapper,
    F,
    OuterRef,
    Q,
    QuerySet,
    Subquery,
    Value,
    When,
)
from django.db.models.functions import Coalesce
from skfolio.preprocessing import prices_to_returns
from wbcore.contrib.currency.models import Currency, CurrencyFXRates

from wbfdm.enums import MarketData

logger = logging.getLogger("pms")


class InstrumentQuerySet(QuerySet):
    def filter_active_at_date(self, val_date: date):
        return self.filter(
            (Q(delisted_date__isnull=True) | Q(delisted_date__gte=val_date))
            & (Q(inception_date__isnull=True) | Q(inception_date__lte=val_date))
        )

    def annotate_classification_for_group(
        self, classification_group, classification_height: int = 0, **kwargs
    ) -> QuerySet:
        return classification_group.annotate_queryset(
            self, classification_height, "", annotation_label="ancestor_classifications", **kwargs
        )

    def annotate_base_data(self):
        base_qs = InstrumentQuerySet(self.model, using=self._db)
        return self.annotate(
            is_investable=~Exists(base_qs.filter(parent=OuterRef("pk"))),
            root=Subquery(base_qs.filter(tree_id=OuterRef("tree_id"), level=0).values("id")[:1]),
            primary_security=ExpressionWrapper(
                Coalesce(
                    Subquery(
                        base_qs.filter(
                            parent=OuterRef("pk"),
                            is_primary=True,
                            is_security=True,
                        ).values("id")[:1]
                    ),
                    F("id"),
                ),
                output_field=AutoField(),
            ),
            _primary_quote=ExpressionWrapper(
                Subquery(
                    base_qs.filter(
                        parent=OuterRef("primary_security"),
                        is_primary=True,
                    ).values("id")[:1]
                ),
                output_field=AutoField(),
            ),
            _preferred_primary_quote=ExpressionWrapper(
                Subquery(
                    base_qs.filter(
                        parent=OuterRef("primary_security"),
                        is_preferred_primary=True,
                    ).values("id")[:1]
                ),
                output_field=AutoField(),
            ),
            primary_quote=Case(
                When(_preferred_primary_quote__isnull=False, then=F("_preferred_primary_quote")),
                When(_primary_quote__isnull=False, then=F("_primary_quote")),
                default=F("primary_security"),
            ),
        )

    def annotate_all(self):
        return self.annotate_base_data()

    @property
    def dl(self):
        """Provides access to the dataloader proxy for the entities in the QuerySet.

        This method allows for easy retrieval of the DataloaderProxy instance
        associated with the QuerySet. It enables the utilization of dataloader
        functionalities directly from the QuerySet, facilitating data fetching and
        processing tasks.

        Returns:
            DataloaderProxy: An instance of DataloaderProxy associated with the
                entities in the QuerySet.
        """
        return self.model.dl_proxy(self)

    def get_instrument_prices_from_market_data(self, from_date: date, to_date: date):
        from wbfdm.models import InstrumentPrice

        def _dict_to_object(instrument, row):
            close = row.get("close", None)
            price_date = row.get("date")
            if price_date and close is not None:
                close = round(Decimal(close), 6)
                # we validate that close can be inserting into our table<
                with suppress(ValidationError):
                    validator = DecimalValidator(16, 6)
                    validator(close)
                    try:
                        try:
                            InstrumentPrice.objects.get(instrument=instrument, date=price_date)
                        except MultipleObjectsReturned:
                            InstrumentPrice.objects.get(
                                instrument=instrument, date=price_date, calculated=False
                            ).delete()
                        p = InstrumentPrice.objects.get(instrument=instrument, date=price_date)
                        p.net_value = close
                        p.gross_value = close
                        p.calculated = row["calculated"]
                        p.volume = row.get("volume", p.volume)
                        p.market_capitalization = row.get("market_capitalization", p.market_capitalization)
                        p.market_capitalization_consolidated = p.market_capitalization
                        p.set_dynamic_field(False)
                        p.id = None
                        return p
                    except InstrumentPrice.DoesNotExist:
                        with suppress(CurrencyFXRates.DoesNotExist):
                            p = InstrumentPrice(
                                currency_fx_rate_to_usd=CurrencyFXRates.objects.get(
                                    # we need to get the currency rate because we bulk create the object, and thus save is not called
                                    date=price_date,
                                    currency=instrument.currency,
                                ),
                                instrument=instrument,
                                date=price_date,
                                calculated=row["calculated"],
                                net_value=close,
                                gross_value=close,
                                volume=row.get("volume", None),
                                market_capitalization=row.get("market_capitalization", None),
                            )
                            p.set_dynamic_field(False)
                            return p

        df = pd.DataFrame(
            self.dl.market_data(
                from_date=from_date,
                to_date=to_date,
                values=[MarketData.CLOSE, MarketData.VOLUME, MarketData.MARKET_CAPITALIZATION],
            )
        )
        if not df.empty:
            df["valuation_date"] = pd.to_datetime(df["valuation_date"])
            df = df.rename(columns={"valuation_date": "date"})
            df = df.drop(
                columns=df.columns.difference(
                    ["calculated", "close", "market_capitalization", "volume", "instrument_id", "date"]
                )
            )
            df["calculated"] = False

            for instrument_id, dff in df.groupby("instrument_id", group_keys=False, as_index=False):
                dff = dff.drop(columns=["instrument_id"]).set_index("date").sort_index()
                if dff.index.duplicated().any():
                    dff = dff.groupby(level=0).first()
                    logger.warning(
                        f"We detected a duplicated index for instrument id {instrument_id}. Please correct the dl parameter which likely introduced this issue."
                    )

                dff = dff.reindex(pd.date_range(dff.index.min(), dff.index.max(), freq="B"))

                dff[["close", "market_capitalization"]] = (
                    dff[["close", "market_capitalization"]].astype(float).ffill(limit_area="inside").ffill(limit=2)
                )
                dff.volume = dff.volume.astype(float).fillna(0)
                dff.calculated = dff.calculated.astype(bool).fillna(
                    True
                )  # we do not ffill calculated but set the to True to mark them as "estimated"/"not real"

                dff = dff.reset_index(names="date").dropna(subset=["close"])
                dff = dff.replace([np.inf, -np.inf, np.nan], None)
                instrument = self.get(id=instrument_id)

                yield from filter(
                    lambda x: x, map(lambda row: _dict_to_object(instrument, row), dff.to_dict("records"))
                )

    def get_returns_df(
        self,
        from_date: date,
        to_date: date,
        to_currency: Currency | None = None,
        use_dl: bool = False,
        ffill_period: int = 10,
    ) -> tuple[dict[date, dict[int, float]], pd.DataFrame]:
        """
        Utility methods to get instrument returns for a given date range

        Args:
            from_date: date range lower bound
            to_date: date range upper bound
            to_currency: currency to use for returns
            use_dl: whether to get data straight from the dataloader or use the internal table

        Returns:
            Return a tuple of the raw prices and the returns dataframe
        """
        padded_from_date = from_date - timedelta(days=15)
        padded_to_date = to_date + timedelta(days=3)
        logger.info(
            f"Loading returns from {from_date:%Y-%m-%d} (padded to {padded_from_date:%Y-%m-%d}) to {to_date:%Y-%m-%d} (padded to {padded_to_date:%Y-%m-%d}) for {self.count()} instruments"
        )

        if use_dl:
            kwargs = dict(
                from_date=padded_from_date, to_date=padded_to_date, values=[MarketData.CLOSE], apply_fx_rate=False
            )
            if to_currency:
                kwargs["target_currency"] = to_currency.key
            df = pd.DataFrame(self.dl.market_data(**kwargs))
            if df.empty:
                df = pd.DataFrame(columns=["instrument_id", "fx_rate", "close", "valuation_date"])
            else:
                df = df[df.columns.intersection(["instrument_id", "fx_rate", "close", "valuation_date"])]
            if "fx_rate" not in df.columns:
                df["fx_rate"] = 1.0
        else:
            from wbfdm.models import InstrumentPrice

            if to_currency:
                fx_rate = Coalesce(
                    CurrencyFXRates.get_fx_rates_subquery_for_two_currencies(
                        "date", "instrument__currency", to_currency
                    ),
                    Decimal("1"),
                )
            else:
                fx_rate = Value(Decimal("1"))
            # annotate fx rate only if the price is not calculated, in that case we assume the instrument is not tradable and we set a forex of None (to be fast forward filled)
            prices = InstrumentPrice.objects.filter(
                instrument__in=self, date__gte=padded_from_date, date__lte=padded_to_date
            ).annotate(fx_rate=Case(When(calculated=False, then=fx_rate), default=None))
            df = pd.DataFrame(
                prices.filter_only_valid_prices().values_list("instrument", "fx_rate", "net_value", "date"),
                columns=["instrument_id", "fx_rate", "close", "valuation_date"],
            )
        df = (
            df.pivot_table(index="valuation_date", columns="instrument_id", values=["fx_rate", "close"], dropna=False)
            .astype(float)
            .sort_index()
        )
        delisted_date_df = pd.DataFrame(
            self.filter(delisted_date__isnull=False).values_list("id", "delisted_date"),
            columns=["id", "delisted_date"],
        ).set_index("id")["delisted_date"]

        if not df.empty:
            ts = pd.bdate_range(df.index.min(), df.index.max(), freq="B")
            df = df.reindex(ts)

            # we allow 10 market days (arbitrary) to be fast forwarded to account for market holiday. This obviously will lead to position being created even though instrument is delisted if its delisted date is not properly set, we have to leave with it.
            df = df.ffill(limit_area="inside").ffill(limit=ffill_period)
            df.index = pd.to_datetime(df.index)
            df = df[
                (df.index <= pd.Timestamp(to_date)) & (df.index >= pd.Timestamp(from_date))
            ]  # ensure the returned df corresponds to requested date range
            prices_df = df["close"]
            delisted_date_df = pd.to_datetime(delisted_date_df)
            delisted_date_df = delisted_date_df.reindex(prices_df.columns).fillna(df.index.max())
            # NumPy broadcast: dates[:, None] <= delist[None, :] => 2D mask (keep if <= delist)
            dates_np = np.array(prices_df.index)
            mask = dates_np[:, np.newaxis] <= np.array(delisted_date_df)
            mask_df = pd.DataFrame(mask, index=df.index, columns=prices_df.columns)
            prices_df = prices_df.where(mask_df, None)
            if "fx_rate" in df.columns:
                fx_rate_df = df["fx_rate"].fillna(1.0)
            else:
                fx_rate_df = pd.DataFrame(np.ones(prices_df.shape), index=prices_df.index, columns=prices_df.columns)
            returns = prices_to_returns(fx_rate_df * prices_df, drop_inceptions_nan=False, fill_nan=True)

            return {
                ts.date(): row for ts, row in prices_df.replace([np.nan], None).to_dict("index").items()
            }, returns.replace([np.inf, -np.inf, np.nan], 0)
        return {}, pd.DataFrame()



from datetime import date
from enum import Enum
from typing import Iterator

from django.db import connections
from jinjasql import JinjaSql  # type: ignore
from wbcore.contrib.dataloader.dataloaders import Dataloader
from wbcore.contrib.dataloader.utils import dictfetchall

from wbfdm.contrib.qa.dataloaders.fx_rates import FXRateConverter
from wbfdm.dataloaders.protocols import StatementsProtocol
from wbfdm.dataloaders.types import StatementDataDict
from wbfdm.enums import DataType, Financial, PeriodType, StatementType


class RKDStatementType(Enum):
    INCOME_STATEMENT = 1
    CASHFLOW_STATEMENT = 2
    BALANCE_SHEET = 3


class RKDFinancial(Enum):
    EMPLOYEES = "METL"
    CASH_AND_SHORT_TERM_INVESTMENT = "SCSI"
    DILUTED_WEIGHTED_AVG_SHARES = "SDWS"
    TOTAL_DEBT = "STLD"
    NET_DEBT = "SNTD"
    STOCK_COMPENSATION = "VSCP"
    TANGIBLE_BOOK_VALUE_PER_SHARE = "STBP"
    REVENUE = "RTLR"
    SHARES_OUTSTANDING = "QTCO"
    EPS = "SDAI"
    CASH_FLOW_FROM_OPERATIONS = "OTLO"
    CAPEX = "SCEX"


reported_sql = """
    SELECT
        CONCAT(val.Code, '-', CONVERT(DATE, val.PerEndDt), '-', fil.InterimNo, '-', val.LineID) as id,
        val.LineID as external_ordering,
        itemsub.Desc_ as external_description,
        itemsub.StdCOA as external_code,
        val.Code as external_identifier,
        CONVERT(DATE, val.PerEndDt) as period_end_date,
        per.Fyr as year,
        fil.InterimNo as interim,
        CASE
            when fil.PerTypeCode = 1 THEN 'Y'
            when fil.PerTypeCode = 2 THEN 'Q'
            when fil.PerTypeCode = 3 THEN 'S'
            when fil.PerTypeCode = 4 THEN 'T'
            when fil.PerTypeCode = 5 THEN 'Q'
            when fil.PerTypeCode = 6 THEN 'Y'
        END as period_type,
        CASE
            WHEN fil.UnitsConvToCode = 'T' AND item.ItemPrecision in (1,2) THEN val.Value_ * 1e3
            WHEN fil.UnitsConvToCode = 'M' AND item.ItemPrecision in (1,2) THEN val.Value_ * 1e6
            WHEN fil.UnitsConvToCode = 'B' AND item.ItemPrecision in (1,2) THEN val.Value_ * 1e9
            ELSE val.Value_
        END as value,
        'qa-rkd' as source,
        CASE
            WHEN item.IsCurrency = 1 THEN code.Desc_
            ELSE NULL
        END AS currency

    FROM RKDFndCSFFinVal AS val

    JOIN RKDFndCSFStmt AS stmt ON
        val.Code = stmt.Code
        AND val.PerEndDt = stmt.PerEndDt
        AND val.PerTypeCode = stmt.PerTypeCode
        AND val.StmtDt = stmt.StmtDt
        AND val.StmtTypeCode = stmt.StmtTypeCode
        AND stmt.CompStmtCode = 1

    JOIN RKDFndCSFPerFiling AS fil ON
        stmt.Code = fil.Code
        AND stmt.PerEndDt = fil.PerEndDt
        AND stmt.PerTypeCode = fil.PerTypeCode
        AND stmt.StmtDt = fil.StmtDt

    LEFT JOIN RKDFNDCSFITEMSUB AS itemsub ON
        val.Code = itemsub.Code
        AND val.PerTypeCode = itemsub.PerTypeCode
        AND val.StmtTypeCode = itemsub.StmtTypeCode
        AND val.LineID = itemsub.LineID

    LEFT JOIN RKDFndCSFItem AS item ON
        itemsub.Item = item.Item

    LEFT JOIN RKDFndCode AS code ON
        fil.CurrConvToCode = code.Code
        AND code.Type_ = 58

    LEFT JOIN RKDFndCsfPeriod AS per ON
        per.Code = fil.Code
        AND per.PerEndDt = fil.PerEndDt
        AND per.PerTypeCode = (
            CASE
                WHEN fil.PerTypeCode = 1 THEN 1
                WHEN fil.PerTypeCode in (2,3,4,5) THEN 5
            END
        )

    WHERE
        val.Code in (
        {% for instrument in instruments %}
            {{instrument}} {% if not loop.last %}, {% endif %}
        {% endfor %})
        AND val.StmtTypeCode = {{ statement_type }}
        {% if from_year %}AND per.Fyr >= {{ from_year }} {% endif %}
        {% if to_year %}AND per.Fyr <= {{ to_year }} {% endif %}
        {% if from_date %}AND val.PerEndDt >= {{ from_date }} {% endif %}
        {% if to_date %}AND val.PerEndDt <= {{ to_date }} {% endif %}
        {% if period_type == 'interim' %}AND fil.InterimNo > 0{% elif period_type == 'annual' %}AND (fil.InterimNo = 0 OR fil.InterimNo IS NULL){% endif %}
    ORDER BY val.LineID
"""
standardized_sql = """
    SELECT
        CONCAT(val.Code, '-', CONVERT(DATE, val.PerEndDt), '-', fil.InterimNo, '-', item.LineID) as id,
        item.LineID as external_ordering,
        item.Desc_ as external_description,
        item.COA as external_code,
        val.Code as external_identifier,
        CONVERT(DATE, val.PerEndDt) as period_end_date,
        per.Fyr as year,
        CASE
            when fil.PerTypeCode = 1 THEN 'Y'
            when fil.PerTypeCode = 2 THEN 'Q'
            when fil.PerTypeCode = 3 THEN 'S'
            when fil.PerTypeCode = 4 THEN 'T'
            when fil.PerTypeCode = 5 THEN 'Q'
            when fil.PerTypeCode = 6 THEN 'Y'
        END as period_type,
        fil.InterimNo as interim,
        CASE
            WHEN fil.UnitsConvToCode = 'T' AND item.ItemPrecision in (1,2) THEN val.Value_ * 1e3
            WHEN fil.UnitsConvToCode = 'M' AND item.ItemPrecision in (1,2) THEN val.Value_ * 1e6
            WHEN fil.UnitsConvToCode = 'B' AND item.ItemPrecision in (1,2) THEN val.Value_ * 1e9
            ELSE val.Value_
        END as value,
        'qa-rkd' as source,
        CASE
            WHEN item.IsCurrency = 1 THEN code.Desc_
            ELSE NULL
        END AS currency

    FROM RKDFndStdFinVal AS val

    LEFT JOIN RKDFndStdPeriod AS per ON
        per.Code = val.Code
        AND per.PerEndDt = val.PerEndDt
        AND per.PerTypeCode = (
            CASE
                WHEN val.PerTypeCode = 1 THEN 1
                WHEN val.PerTypeCode in (2,3,4,5) THEN 5
            END
        )
        AND per.PerEnddt = (
            SELECT TOP 1 per2.PerEndDt
            FROM RKDFndStdPeriod AS per2
            WHERE per.Code = per2.Code
              AND per.PerTypeCode = per2.PerTypeCode
              AND per.Fyr = per2.Fyr
              AND (
                per2.InterimNo = per.InterimNo
                    OR (
                    per2.InterimNo IS NULL AND per.InterimNo IS NULL
                    )
                )
            ORDER BY per2.PerEndDt DESC
        )

    JOIN RKDFndStdStmt AS stmt ON
        val.Code = stmt.Code
        AND val.PerEndDt = stmt.PerEndDt
        AND val.PerTypeCode = stmt.PerTypeCode
        AND val.StmtDt = stmt.StmtDt
        AND val.StmtTypeCode = stmt.StmtTypeCode
        AND stmt.CompStmtCode = 1
        AND stmt.StmtDt = (
            select top 1 stmt2.StmtDt
            from RKDFndStdStmt as stmt2
            where stmt2.Code = stmt.Code
                and stmt2.PerEndDt = stmt.PerEndDt
                and stmt2.PerTypeCode = stmt.PerTypeCode
                and stmt2.StmtTypeCode = stmt.StmtTypeCode
                and stmt2.CompStmtCode = stmt.CompStmtCode
            order by stmt2.StmtDt desc
        )

    JOIN RKDFndStdPerFiling AS fil ON
        stmt.Code = fil.Code
        AND stmt.PerEndDt = fil.PerEndDt
        AND stmt.PerTypeCode = fil.PerTypeCode
        AND stmt.StmtDt = fil.StmtDt

    LEFT JOIN RKDFndStdItem AS item ON
        val.Item = item.Item

    LEFT JOIN RKDFndCode AS code ON
        fil.CurrConvToCode = code.Code
        AND code.Type_ = 58

    WHERE
        val.Code in (
        {% for instrument in instruments %}
            {{instrument}} {% if not loop.last %}, {% endif %}
        {% endfor %})
        {% if statement_type %} AND val.StmtTypeCode = {{ statement_type }}{% endif %}
        {% if from_year %}AND per.Fyr >= {{ from_year }} {% endif %}
        {% if to_year %}AND per.Fyr <= {{ to_year }} {% endif %}
        {% if from_date %}AND val.PerEndDt >= {{ from_date }} {% endif %}
        {% if to_date %}AND val.PerEndDt <= {{ to_date }} {% endif %}
        {% if period_type == 'interim' %}AND fil.InterimNo > 0{% elif period_type == 'annual' %}AND (fil.InterimNo = 0 OR fil.InterimNo IS NULL){% endif %}
        {% if external_codes %}
            AND item.COA in (
                {% for external_code in external_codes %}
                    {{ external_code }} {% if not loop.last %}, {% endif %}
                {% endfor %}
            )
        {% endif %}
    ORDER BY item.LineID
"""


class RKDStatementsDataloader(FXRateConverter, StatementsProtocol, Dataloader):
    def statements(
        self,
        statement_type: StatementType | None = None,
        from_date: date | None = None,
        to_date: date | None = None,
        from_year: int | None = None,
        to_year: int | None = None,
        period_type: PeriodType = PeriodType.ALL,
        data_type: DataType = DataType.STANDARDIZED,
        financials: list[Financial] | None = None,
        target_currency: str | None = None,
    ) -> Iterator[StatementDataDict]:
        lookup = {k: v for k, v in self.entities.values_list("dl_parameters__statements__parameters", "id")}
        sql = reported_sql if data_type is DataType.REPORTED else standardized_sql
        if not financials:
            financials = []
        external_codes = [RKDFinancial[fin.name].value for fin in financials if fin.name in RKDFinancial.__members__]
        if from_year and not from_date:
            from_date = date(year=from_year, month=1, day=1)
        if to_date and not to_date:
            to_date = date(year=to_year + 1, month=1, day=1)
        if target_currency:
            self.load_fx_rates(self.entities, target_currency, from_date, to_date)
        query, bind_params = JinjaSql(param_style="format").prepare_query(
            sql,
            {
                "instruments": lookup.keys(),
                "statement_type": RKDStatementType[statement_type.name].value if statement_type else None,
                "from_year": from_year,
                "to_year": to_year,
                "from_date": from_date,
                "to_date": to_date,
                "period_type": period_type.value,
                "external_codes": external_codes,
            },
        )
        with connections["qa"].cursor() as cursor:
            cursor.execute(
                query,
                bind_params,
            )
            for row in dictfetchall(cursor):
                if row["interim"] is None:
                    row["interim"] = 0
                # sometime we get None for the year. We default to the period end date year then
                row["year"] = int(row["year"] or row["period_end_date"].year)
                row["instrument_id"] = lookup[row["external_identifier"]]
                if financials:
                    try:
                        row["financial"] = Financial[RKDFinancial(row["external_code"]).name].value
                    except (ValueError, KeyError):
                        continue
                    if target_currency:
                        row = self.apply_fx_rate(row, ["value"], apply=True, date_label="period_end_date")
                yield row



import logging
from typing import Any

from celery import shared_task
from django.contrib.contenttypes.models import ContentType
from langchain_core.messages import HumanMessage, SystemMessage
from pydantic import BaseModel, Field, ValidationError
from wbcore.contrib.ai.exceptions import APIStatusErrors, BadRequestErrors
from wbcore.contrib.ai.llm.utils import run_llm
from wbcore.workers import Queue

logger = logging.getLogger("llm")


class CompanyModel(BaseModel):
    name: str = Field(..., description="The name of the publicly listed company.")
    isin: str | None = Field(..., description="The ISIN of the publicly listed company. May be null if not available.")
    ticker: str | None = Field(
        ..., description="The ticker of the publicly listed company. May be null if not available."
    )
    refinitiv_identifier_code: str | None = Field(
        ..., description="The RIC of the publicly listed company. May be null if not available."
    )
    sentiment: int = Field(
        ...,
        ge=1,
        le=4,
        description="The sentiment of the news article towards the company. 1 is negative, 2 is neutral, 3 is positive, 4 is very positive.",
    )
    analysis: str = Field(
        ...,
        description="The analysis of the news article towards the company. This should include an explanation of the sentiment.",
    )


class CompaniesModel(BaseModel):
    companies: list[CompanyModel] = Field(
        default=[], description="The list of companies mentioned in the news article."
    )


@shared_task(
    queue=Queue.BACKGROUND.value,
    autoretry_for=tuple(APIStatusErrors),
    retry_backoff=10,
    max_retries=5,  # retry i5 times maximum
    default_retry_delay=30,  # retry in 30s
    retry_jitter=False,
)
def run_company_extraction_llm(title: str, description: str, *args) -> list[dict[str, Any]]:
    from wbfdm.import_export.handlers.instrument import InstrumentLookup
    from wbfdm.models import Instrument

    relationships = []

    try:
        res = run_llm(
            prompt=[
                SystemMessage(
                    content="You will be parsed a news article, please provide the name of the publicly listed companies mentioned in the article, along with their ISIN, ticker, RIC, sentiment, and analysis."
                ),
                HumanMessage(content="Title: {title}, Description: {description}"),
            ],
            output_model=CompaniesModel,
            query={"title": title, "description": description},
        )[0]
        if isinstance(res, CompaniesModel):
            instrument_ct = ContentType.objects.get_for_model(Instrument)
            for company in res.companies:
                instrument = InstrumentLookup(Instrument).lookup(
                    only_security=True,
                    name=company.name,
                    isin=company.isin,
                    ticker=company.ticker,
                    refinitiv_identifier_code=company.refinitiv_identifier_code,
                )
                if instrument is not None:
                    relationships.append(
                        {
                            "content_type_id": instrument_ct.id,
                            "object_id": instrument.get_root().id,
                            "sentiment": company.sentiment,
                            "analysis": company.analysis,
                            "content_object_repr": str(instrument),
                        }
                    )
    except (
        ValidationError,
        *BadRequestErrors,
    ):  # we silent bad request error because there is nothing we can do about it
        pass
    except tuple(APIStatusErrors) as e:  # for APIStatusError, we let celery retry it
        raise e
    except Exception as e:  # otherwise we log the error and silently fail
        logger.warning(str(e))
    return relationships



import pandas as pd
from django.utils.functional import cached_property
from rest_framework.exceptions import ParseError
from wbcore.contrib.io.viewsets import ExportPandasAPIViewSet
from wbcore.contrib.pandas import fields as pf
from wbcore.contrib.pandas.utils import override_number_to_percent
from wbcore.serializers.fields.types import DisplayMode

from wbfdm.analysis.financial_analysis.financial_metric_analysis import (
    financial_metric_estimate_analysis,
    financial_metric_growths,
)
from wbfdm.enums import Financial
from wbfdm.filters import GroupKeyFinancialsFilterSet
from wbfdm.models.instruments import Instrument
from wbfdm.viewsets.configs.display.statement_with_estimates import (
    StatementWithEstimatesDisplayViewConfig,
)
from wbfdm.viewsets.configs.endpoints.statements import StatementsEndpointViewConfig
from wbfdm.viewsets.configs.titles.statement_with_estimates import (
    StatementTitleViewConfig,
)

from ..mixins import InstrumentMixin


class FinancialMetricAnalysisPandasViewSet(InstrumentMixin, ExportPandasAPIViewSet):
    queryset = Instrument.objects.none()
    display_config_class = StatementWithEstimatesDisplayViewConfig
    endpoint_config_class = StatementsEndpointViewConfig
    title_config_class = StatementTitleViewConfig
    filterset_class = GroupKeyFinancialsFilterSet

    def get_queryset(self):
        return Instrument.objects.filter(id=self.instrument.id)

    def get_pandas_fields(self, request):
        return pf.PandasFields(
            fields=[
                *[pf.FloatField(key=field, label=field, display_mode=DisplayMode.SHORTENED) for field in self.columns],
                pf.PKField(key="id", label="ID"),
                pf.CharField(key="financial", label="Financial"),
                pf.JsonField(key="_overwrites", label="Overwrites"),
            ]
        )

    def get_dataframe(self, request, queryset, **kwargs):
        if group_keys := request.GET.get("group_keys"):
            try:
                financial = Financial(group_keys.lower())
            except ValueError as e:
                raise ParseError() from e
            df, self._estimate_mapping, self._columns = financial_metric_estimate_analysis(
                queryset.first().id, financial
            )
            empty_row = pd.Series([None], dtype="float64", name="empty_row")
            df_growth = financial_metric_growths(queryset.first().id, financial)

            df = pd.concat([df_growth, df, empty_row]).dropna(how="all")
            if "financial" in df.columns:
                override_number_to_percent(df, df["financial"].str.contains("(%)"))

            df = df.rename(columns={"index": "id"})
            return df

        return pd.DataFrame()

    @cached_property
    def columns(self):
        if not hasattr(self, "_columns"):
            self.get_dataframe(self.request, self.get_queryset())
        return self._columns

    @property
    def year_columns(self):
        yield from filter(lambda col: "Y" in col, self.columns)

    @property
    def interim_columns(self):
        yield from filter(lambda col: "Y" not in col, self.columns)

    @cached_property
    def estimate_mapping(self):
        if not hasattr(self, "_estimate_mapping"):
            self.get_dataframe(self.request, self.get_queryset())
        return self._estimate_mapping



from rest_framework.reverse import reverse
from wbcore.metadata.configs.endpoints import EndpointViewConfig


class RelatedInstrumentThroughInstrumentEndpointConfig(EndpointViewConfig):
    def get_endpoint(self, **kwargs):
        return reverse(
            "wbfdm:instrument-relatedinstrument-list",
            args=[self.view.kwargs["instrument_id"]],
            request=self.request,
        )


class ClassifiedInstrumentEndpointConfig(EndpointViewConfig):
    PK_FIELD = "instrument"

    def get_endpoint(self, **kwargs):
        return reverse("wbfdm:classifiedinstrument-list", args=[], request=self.request)

    def get_instance_endpoint(self, **kwargs):
        return reverse(
            "wbfdm:instrument-list",
            args=[],
            request=self.request,
        )

    def get_create_endpoint(self, **kwargs):
        return None

    def get_delete_endpoint(self, **kwargs):
        return None


class InstrumentFavoriteGroupEndpointConfig(EndpointViewConfig):
    def get_update_endpoint(self, **kwargs):
        if self.view.is_owner:
            return self.get_instance_endpoint(**kwargs)
        return None

    def get_delete_endpoint(self, **kwargs):
        return self.get_update_endpoint(**kwargs)



from datetime import date, timedelta

import pandas as pd
from wbcore.contrib.io.viewsets import ExportPandasAPIViewSet
from wbcore.contrib.pandas import fields as pf
from wbcore.serializers.fields.types import DisplayMode
from wbcore.utils.date import get_date_interval_from_request

from wbfdm.filters.instrument_prices import FakeDateRange
from wbfdm.models.instruments import Instrument
from wbfdm.viewsets.configs.display.prices import InstrumentPriceDisplayConfig
from wbfdm.viewsets.configs.titles.prices import InstrumentPriceTitleViewConfig

from .mixins import InstrumentMixin


class InstrumentPriceViewSet(InstrumentMixin, ExportPandasAPIViewSet):
    IDENTIFIER = "wbfdm:instrument-price"
    display_config_class = InstrumentPriceDisplayConfig
    title_config_class = InstrumentPriceTitleViewConfig
    pandas_fields = pf.PandasFields(
        fields=[
            pf.PKField(key="id", label="ID"),
            pf.DateField(
                key="valuation_date",
                label="valuation_date",
            ),
            pf.FloatField(key="open", label="open"),
            pf.FloatField(key="high", label="high"),
            pf.FloatField(key="low", label="low"),
            pf.FloatField(key="close", label="close"),
            pf.FloatField(key="volume", label="volume", display_mode=DisplayMode.SHORTENED),
            pf.FloatField(key="outstanding_shares", label="outstanding_shares", display_mode=DisplayMode.SHORTENED),
            pf.FloatField(
                key="market_capitalization", label="market_capitalization", display_mode=DisplayMode.SHORTENED
            ),
            pf.FloatField(
                key="market_capitalization_consolidated",
                label="market_capitalization_consolidated",
                display_mode=DisplayMode.SHORTENED,
            ),
        ]
    )
    permission_classes = []
    filterset_class = FakeDateRange
    queryset = Instrument.objects.all()
    ordering_fields = (
        "valuation_date",
        "open",
        "high",
        "low",
        "close",
        "volume",
        "outstanding_shares",
        "market_capitalization",
    )
    ordering = ["-valuation_date"]

    def get_queryset(self):
        return Instrument.objects.filter(id=self.instrument.id)

    def get_dataframe(self, request, queryset, **kwargs):
        start, end = get_date_interval_from_request(request, date_range_fieldname="date")
        if not end:
            end = date.today()
        if not start:
            start = end - timedelta(days=365)
        return pd.DataFrame(queryset.dl.market_data(from_date=start, to_date=end))



from wbcore.menus import ItemPermission, MenuItem

INSTRUMENTMETRIC_MENUITEM = MenuItem(
    label="Metrics",
    endpoint="metric:instrumentmetric-list",
    endpoint_get_parameters={"parent_metric__isnull": True},
    permission=ItemPermission(
        permissions=["metric.view_instrumentmetric"], method=lambda request: request.user.is_internal
    ),
)



from collections import defaultdict
from datetime import date, timedelta
from typing import Iterator

import pypika as pk
from django.db import connections
from pypika import Case
from pypika import functions as fn
from pypika.enums import Order, SqlTypes
from wbcore.contrib.dataloader.dataloaders import Dataloader
from wbcore.contrib.dataloader.utils import dictfetchall

from wbfdm.dataloaders.protocols import FXRateProtocol
from wbfdm.dataloaders.types import FXRateDict


class FXRateConverter:
    fx_rates: dict[str, dict[date, float]]
    target_currency: str

    def load_fx_rates(self, entities, target_currency: str, from_date: date, to_date: date):
        fx_rates = defaultdict(dict)
        self.target_currency = target_currency
        if from_date:
            for fx_rate in DatastreamFXRatesDataloader(entities).fx_rates(
                from_date, to_date if to_date else date.today(), target_currency
            ):
                fx_rates[fx_rate["currency_pair"]][fx_rate["fx_date"]] = fx_rate["fx_rate"]
        self.fx_rates = fx_rates

    def apply_fx_rate(self, row, currency_columns: list[str], apply: bool = True, date_label: str = "valuation_date"):
        if row["currency"]:
            fx_rate = None
            if self.target_currency == row["currency"]:
                fx_rate = 1.0
            else:
                currency_fx_rates = self.fx_rates[f'{row["currency"]}{self.target_currency}']
                if currency_fx_rates:
                    try:
                        fx_rate = currency_fx_rates[row[date_label]] or 1.0
                    except KeyError:
                        max_idx = max(currency_fx_rates.keys())
                        fx_rate = currency_fx_rates[max_idx]
            if apply and fx_rate is not None:
                for col in currency_columns:
                    if v := row.get(col):
                        row[col] = v * fx_rate
                row["currency"] = self.target_currency
            row["fx_rate"] = fx_rate
        return row


class DatastreamFXRatesDataloader(FXRateProtocol, Dataloader):
    def fx_rates(
        self,
        from_date: date,
        to_date: date,
        target_currency: str,
    ) -> Iterator[FXRateDict]:
        currencies = list(self.entities.values_list("currency__key", flat=True))
        # Define tables
        fx_rate = pk.Table("DS2FxRate")
        fx_code = pk.Table("DS2FxCode")

        # Base query to get data we always need unconditionally
        query = (
            pk.MSSQLQuery.from_(fx_rate)
            # We join on _codes, which removes all instruments not in _codes - implicit where
            .join(fx_code)
            .on(fx_rate.ExRateIntCode == fx_code.ExRateIntCode)
            .where((fx_rate.ExRateDate >= from_date) & (fx_rate.ExRateDate <= to_date + timedelta(days=1)))
            .where(
                (fx_code.ToCurrCode == target_currency)
                & (fx_code.FromCurrCode.isin(currencies))
                & (fx_code.RateTypeCode == "SPOT")
            )
            .orderby(fx_rate.ExRateDate, order=Order.desc)
            .select(
                fn.Cast(fx_rate.ExRateDate, SqlTypes.DATE).as_("fx_date"),
                fn.Concat(fx_code.FromCurrCode, fx_code.ToCurrCode).as_("currency_pair"),
                (Case().when(fx_code.FromCurrCode == target_currency, 1).else_(1 / fx_rate.midrate)).as_("fx_rate"),
            )
        )
        with connections["qa"].cursor() as cursor:
            cursor.execute(query.get_sql())
            yield from dictfetchall(cursor, FXRateDict)
